__all__ = ()

import reprlib

from scarletio import Compound

from ...allowed_mentions import parse_allowed_mentions
from ...application import Application
from ...bases import maybe_snowflake
from ...channel import Channel, ChannelType, create_partial_channel_from_id
from ...http import DiscordHTTPClient, VALID_ICON_MEDIA_TYPES_EXTENDED
from ...message import Message
from ...utils import get_image_media_type, image_to_base64
from ...webhook import Webhook, create_partial_webhook_from_id

from ..request_helpers import (
    add_file_to_message_data, get_channel_id, get_components_data, get_guild_id, get_webhook_and_id,
    get_webhook_and_id_and_token, get_webhook_id, get_webhook_id_and_token, validate_content_and_embed,
)


class ClientCompoundWebhookEndpoints(Compound):
    
    application : Application
    http : DiscordHTTPClient
    
    
    async def webhook_create(self, channel, name, *, avatar=None):
        """
        Creates a webhook at the given channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel of the created webhook.
        name : `str`
            The name of the new webhook. It's length can be in range [1:80].
        avatar : `None`, `bytes-like` = `None`, Optional (Keyword only)
            The webhook's avatar. Can be `'jpg'`, `'png'`, `'webp'`, `'gif'` image's raw data. However if set as
            `'gif'`, it will not have any animation.
            
        Returns
        -------
        webhook : ``Webhook``
            The created webhook.
        
        Raises
        ------
        TypeError
            - If `channel` was not given neither as ``Channel`` nor as `int`.
            - If `avatar` was not given neither as `None`, `bytes-like`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` was not given as `str`.
            - If `name` range is out of the expected range [1:80].
            - If `avatar`'s type is not any of the expected ones: `'jpg'`, `'png'`, `'webp'`, `'gif'`.
        """
        channel_id = get_channel_id(channel, Channel.is_in_group_guild_system)
        
        if __debug__:
            if not isinstance(name, str):
                raise AssertionError(
                    f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                )
            
            name_length = len(name)
            if name_length < 1 or name_length > 80:
                raise AssertionError(
                    f'`name` length can be in range [1:80], got {name_length!r}; {name!r}.'
                )
        
        data = {'name': name}
        
        if (avatar is not None):
            if not isinstance(avatar, (bytes, bytearray, memoryview)):
                raise TypeError(
                    f'`avatar` can `None`, `bytes-like`, got {avatar.__name__}; {reprlib.repr(avatar)}.'
                )
            
            if __debug__:
                media_type = get_image_media_type(avatar)
                if media_type not in VALID_ICON_MEDIA_TYPES_EXTENDED:
                    raise AssertionError(
                        f'Invalid `avatar` media type: {media_type}; got {reprlib.repr(avatar)}.'
                    )
            
            data['avatar'] = image_to_base64(avatar)
        
        data = await self.http.webhook_create(channel_id, data)
        return Webhook(data)
    
    
    async def webhook_get(self, webhook):
        """
        Requests the webhook by it's id.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `int`
            The webhook to update or the webhook's id to get.
        
        Returns
        -------
        webhook : ``Webhook``
        
        Raises
        ------
        TypeError
            If `webhook` was not given neither as ``Webhook`` neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        ``.webhook_get_token`` : Getting webhook with Discord's webhook API.
        
        Notes
        -----
        If the webhook already loaded and if it's guild's webhooks are up to date, no request is done.
        """
        webhook, webhook_id = get_webhook_and_id(webhook)
        
        data = await self.http.webhook_get(webhook_id)
        if webhook is None:
            webhook = Webhook(data)
        else:
            webhook._update_attributes(data)
        
        return webhook
    
    
    async def webhook_get_token(self, webhook):
        """
        Requests the webhook through Discord's webhook API. The client do not needs to be in the guild of the webhook.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook to update or the webhook's id and token.
        
        Returns
        -------
        webhook : ``Webhook``
        
        Raises
        ------
        TypeError
            If `webhook` was not given neither as ``Webhook`` neither as a `tuple` (`int`, `str`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        If the webhook already loaded and if it's guild's webhooks are up to date, no request is done.
        """
        webhook, webhook_id, webhook_token = get_webhook_and_id_and_token(webhook)
        
        if (webhook is None):
            webhook = create_partial_webhook_from_id(webhook_id, webhook_token)
        
        data = await self.http.webhook_get_token(webhook_id, webhook_token)
        webhook._set_attributes(data)
        return webhook
    
    
    async def webhook_get_all_channel(self, channel):
        """
        Requests all the webhooks of the channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel, what's webhooks will be requested.
        
        Returns
        -------
        webhooks : `list` of ``Webhook` objects
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel``, neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
            
            You may expect the following exceptions:
            
            +---------------+-----------------------+---------------------------------------------------------------+
            | Error code    | Internal name         | Reason                                                        |
            +===============+=======================+===============================================================+
            | 10003         | unknown_channel       | The channel not exists.                                       |
            +---------------+-----------------------+---------------------------------------------------------------+
            | 50013         | missing_permissions   | You need `manage_webhooks` permission. (Or the client has no  |
            |               |                       | access to the channel.)                                       |
            +---------------+-----------------------+---------------------------------------------------------------+
            | 60003         | MFA_required          | You need to have multi-factor authorization to do this        |
            |               |                       | operation (guild setting dependent). For bot accounts it      |
            |               |                       | means their owner needs mfa.                                  |
            +---------------+-----------------------+---------------------------------------------------------------+
            
            > Discord drops `Forbidden (403), code=50013: Missing Permissions` instead of
            > `Forbidden (403), code=50001: Missing Access`.
        
        AssertionError
            If `channel` was given as a channel's identifier but it detectably not refers to a ``Channel``.
        
        Notes
        -----
        No request is done, if the passed channel is partial, or if the channel's guild's webhooks are up to date.
        """
        channel_id = get_channel_id(channel, Channel.is_in_group_guild_system)
        
        data = await self.http.webhook_get_all_channel(channel_id)
        return [Webhook(webhook_data) for webhook_data in data]
    
    
    async def webhook_get_own_channel(self, channel):
        """
        Requests the webhooks of the given channel and returns the first owned one.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel, what's webhooks will be requested.
        
        Returns
        -------
        webhooks : `list` of ``Webhook` objects
        """
        webhooks = await self.webhook_get_all_channel(channel)
        
        application_id = self.application.id
        for webhook in webhooks:
            if webhook.application_id == application_id:
                return webhook
        
        return None
    
    
    async def webhook_get_all_guild(self, guild):
        """
        Requests the webhooks of the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's webhooks will be requested.
        
        Returns
        -------
        webhooks : `list` of ``Webhook` objects
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        No request is done, if the guild's webhooks are up to date.
        """
        guild_id = get_guild_id(guild)
        
        webhook_datas = await self.http.webhook_get_all_guild(guild_id)
        return [Webhook(webhook_data) for webhook_data in webhook_datas]
    
    
    async def webhook_delete(self, webhook):
        """
        Deletes the webhook.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `int`
            The webhook to delete.
        
        Raises
        ------
        TypeError
            If `webhook` was not given neither as ``Webhook``, `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        ``.webhook_delete_token`` : Deleting webhook with Discord's webhook API.
        """
        webhook_id = get_webhook_id(webhook)
        
        await self.http.webhook_delete(webhook_id)
    
    
    async def webhook_delete_token(self, webhook):
        """
        Deletes the webhook through Discord's webhook API.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook to delete.

        Parameters
        ----------
        webhook : ``Webhook``
            The webhook to delete.
        
        Raises
        ------
        TypeError
            If `webhook` was not given neither as ``Webhook`` neither as a `tuple` (`int`, `str`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        webhook_id, webhook_token = get_webhook_id_and_token(webhook)
        
        await self.http.webhook_delete_token(webhook_id, webhook_token)
    
    # later there gonna be more stuff that's why 2 different
    async def webhook_edit(self, webhook, *, name=..., avatar=..., channel = ...):
        """
        Edits and updates the given webhook.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `int`
            The webhook to edit.
        name : `str`, Optional (Keyword only)
            The webhook's new name. It's length can be in range [1:80].
        avatar : `None`, `bytes-like`, Optional (Keyword only)
            The webhook's new avatar. Can be `'jpg'`, `'png'`, `'webp'`, `'gif'` image's raw data. However if set as
            `'gif'`, it will not have any animation. If passed as `None`, will remove the webhook's current avatar.
        channel : ``Channel``, `int`, Optional (Keyword only)
            The webhook's channel.
        
        Raises
        ------
        TypeError
            - If `webhook` was not given neither as ``Webhook`` neither as `int`.
            - If `avatar` was not given neither as `None` nor as `bytes-like`.
            - If `channel` was not given neither as ``Channel`` neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` was given but not as `str`.
            - If `name`'s length is out of range [1:80].
            - If `avatar`'s type is not any of the expected ones: `'jpg'`, `'png'`, `'webp'`, `'gif'`.
        
        See Also
        --------
        ``.webhook_edit_token`` : Editing webhook with Discord's webhook API.
        """
        webhook_id = get_webhook_id(webhook)
        
        data = {}
        
        if (name is not ...):
            if __debug__:
                if not isinstance(name, str):
                    raise AssertionError(
                        f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                    )
                
                name_length = len(name)
                if name_length < 1 or name_length > 80:
                    raise AssertionError(
                        f'The length of the name can be in range [1:80], got {name_length}; {name!r}.'
                    )
            
            data['name'] = name
        
        if (avatar is not ...):
            if avatar is None:
                avatar_data = None
            else:
                if not isinstance(avatar, (bytes, bytearray, memoryview)):
                    raise TypeError(
                        f'`avatar` can be `None`, `bytes-like`, got {avatar.__class__.__name__}; '
                        f'{reprlib.repr(avatar)}.'
                    )
                
                if __debug__:
                    media_type = get_image_media_type(avatar)
                    if media_type not in VALID_ICON_MEDIA_TYPES_EXTENDED:
                        raise AssertionError(
                            f'Invalid `avatar` type for the client: {media_type}; got {reprlib.repr(avatar)}.'
                        )
                
                avatar_data = image_to_base64(avatar)
            
            data['avatar'] = avatar_data
        
        if (channel is not ...):
            while True:
                if isinstance(channel, Channel):
                    if channel.is_in_group_guild_system() or channel.partial:
                        channel_id = channel.id
                        break
                
                else:
                    channel_id = maybe_snowflake(channel)
                    if channel_id is not None:
                        break
                
                raise TypeError(
                    f'`channel` can be gild text channel, `int`, got {channel.__class__.__name__}; '
                    f'{channel!r}.'
                )
            
            data['channel_id'] = channel_id
        
        if not data:
            return # Save 1 request
        
        data = await self.http.webhook_edit(webhook_id, data)
        webhook._set_attributes(data)
    
    
    async def webhook_edit_token(self, webhook, *, name=..., avatar=...):
        """
        Edits and updates the given webhook through Discord's webhook API.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook to edit.
        name : `str`, Optional (Keyword only)
            The webhook's new name. It's length can be between `1` and `80`.
        avatar : `None`, `bytes-like`, Optional (Keyword only)
            The webhook's new avatar. Can be `'jpg'`, `'png'`, `'webp'`, `'gif'` image's raw data. However if set as
            `'gif'`, it will not have any animation. If passed as `None`, will remove the webhook's current avatar.
        
        Returns
        -------
        webhook : ``Webhook``
            The updated webhook.
        
        Raises
        ------
        TypeError
            - If `webhook` was not given neither as ``Webhook`` neither as a `tuple` (`int`, `str`).
            - If `avatar` was not given neither as `None` nor as `bytes-like`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` was given but not as `str`.
            - If `name`'s length is out of range [1:80].
            - If `avatar`'s type is not any of the expected ones: `'jpg'`, `'png'`, `'webp'`, `'gif'`.
        
        Notes
        -----
        This endpoint cannot edit the webhook's channel, like ``.webhook_edit``.
        """
        webhook, webhook_id, webhook_token = get_webhook_and_id_and_token(webhook)
        
        data = {}
        
        if (name is not ...):
            if __debug__:
                if not isinstance(name, str):
                    raise AssertionError(
                        f'`name` can be `None`, `str`, got {name.__class__.__name__}; {name!r}.'
                    )
                
                name_length = len(name)
                if name_length < 1 or name_length > 80:
                    raise AssertionError(
                        f'`name`\'s length can be in range [1:80], got {name_length}; {name!r}.'
                    )
            
            data['name'] = name
        
        if (avatar is not ...):
            if avatar is None:
                avatar_data = None
            else:
                if not isinstance(avatar, (bytes, bytearray, memoryview)):
                    raise TypeError(
                        f'`avatar` can be `None`, `bytes-like`, got {avatar.__class__.__name__}; '
                        f'{reprlib.repr(avatar)}.'
                    )
                
                if __debug__:
                    media_type = get_image_media_type(avatar)
                    if media_type not in VALID_ICON_MEDIA_TYPES_EXTENDED:
                        raise AssertionError(
                            f'Invalid avatar type for the client: {media_type}; got {reprlib.repr(avatar)}.'
                        )
                
                avatar_data = image_to_base64(avatar)
            
            data['avatar'] = avatar_data
        
        if not data:
            return # Save 1 request
        
        data = await self.http.webhook_edit_token(webhook_id, webhook_token, data)
        
        if webhook is None:
            webhook = Webhook(data)
        else:
            webhook._set_attributes(data)
        
        return webhook
    
    
    async def webhook_message_create(
        self, webhook, content=None, *, embed = None, file=None, allowed_mentions = ..., components = None, tts=False,
        name=None, avatar_url = None, thread=None, wait=False
    ):
        """
        Sends a message with the given webhook. If there is nothing to send, or if `wait` was not passed as `True`
        returns `None`.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook through what will the message be sent.
        
        content : `None`, `str`, ``EmbedBase``, `Any` = `None`, Optional
            The message's content if given. If given as `str` or empty string, then no content will be sent, meanwhile
            if any other non `str`, ``EmbedBase`` is given, then will be casted to string.
            
            If given as ``EmbedBase``, then is sent as the message's embed.
            
        embed : `None`, ``EmbedBase``, `list` of ``EmbedBase`` = `None`, Optional (Keyword only)
            The embedded content of the message.
            
            If `embed` and `content` parameters are both given as  ``EmbedBase``, then `AssertionError` is
            raised.
        
        file : `None`, `Any` = `None`, Optional (Keyword only)
            A file or files to send. Check ``create_file_form`` for details.
        
        allowed_mentions : `None`, `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions`` for details.
        
        components : `None`, ``Component``, (`tuple`, `list`) of (``Component``, (`tuple`, `list`) of
                ``Component``) = `None`, Optional (Keyword only)
            Components attached to the message.
            
            > `components` do not count towards having any content in the message.
        
        tts : `bool` = `False`, Optional (Keyword only)
            Whether the message is text-to-speech.
        
        name : `None`, `str` = `None`, Optional (Keyword only)
            The message's author's new name. Default to the webhook's name by Discord.
        
        avatar_url : `None`, `str` = `None`, Optional (Keyword only)
            The message's author's avatar's url. Defaults to the webhook's avatar' url by Discord.
        
        thread : `None`, ``Channel``, `int` = `None`, Optional (Keyword only)
            The thread of the webhook's channel where the message should be sent.
        
        wait : `None`, `bool` = `None`, Optional (Keyword only)
            Whether we should wait for the message to send and receive it's data as well.
        
        Returns
        -------
        message : ``Message``, `None`
            Returns `None` if there is nothing to send or if `wait` was given as `False` (so by default).
        
        Raises
        ------
        TypeError
            - If `webhook` was not given neither as ``Webhook`` neither as a `tuple` (`int`, `str`).
            - If `allowed_mentions` contains an element of invalid type.
            - If `embed` was not given neither as ``EmbedBase`` nor as `list`, `tuple` of ``EmbedBase``-s.
            - `content` parameter was given as ``EmbedBase``, meanwhile `embed` parameter was given as well.
            - If invalid file type would be sent.
            - If `thread` was not given either as `None`, ``Channel`` nor as `int`.
            - If `components` type is incorrect.
        ValueError
            - If `allowed_mentions`'s elements' type is correct, but one of their value is invalid.
            - If more than `10` files would be sent.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` was not passed neither as `None`, `str`.
            - If `name` was passed as `str`, but it's length is out of range [1:32].
            - If `avatar_url` was not given as `str`.
            - If `tts` was not given as `bool`.
            - If `wait` was not given as `bool`.
            - If `embed` contains a non ``EmbedBase`` element.
            - If both `content` and `embed` fields are embeds.
        
        See Also
        --------
        - ``.message_create`` : Create a message with your client.
        - ``.webhook_message_edit`` : Edit a message created by a webhook.
        - ``.webhook_message_delete`` : Delete a message created by a webhook.
        - ``.webhook_message_get`` : Get a message created by a webhook.
        """
        webhook, webhook_id, webhook_token = get_webhook_and_id_and_token(webhook)
        
        while True:
            if thread is None:
                thread_id = 0
                break
                
            elif isinstance(thread, Channel):
                if thread.is_in_group_thread() or thread.partial:
                    thread_id = thread.id
                    break
            
            else:
                thread_id = maybe_snowflake(thread)
                if thread_id is not None:
                    break
            
            raise TypeError(
                f'`thread` can be `None`, thread channel, `int`, '
                f'got {thread.__class__.__name__}; {thread!r}.'
            )
        
        content, embed = validate_content_and_embed(content, embed, False)
        
        components = get_components_data(components, False)
        
        if __debug__:
            if not isinstance(tts, bool):
                raise AssertionError(
                    f'`tts` can be `bool`, got {tts.__class__.__name__}; {tts!r}.'
                )
            
            if not isinstance(wait, bool):
                raise AssertionError(
                    f'`wait` can be `bool`, got {wait.__class__.__name__}; {wait!r}.'
                )
        
        message_data = {}
        contains_content = False
        
        if (content is not None):
            message_data['content'] = content
            contains_content = True
        
        if (embed is not None):
            message_data['embeds'] = [embed.to_data() for embed in embed]
            contains_content = True
        
        if (allowed_mentions is not ...):
            message_data['allowed_mentions'] = parse_allowed_mentions(allowed_mentions)
        
        if (components is not None):
            message_data['components'] = components
        
        if tts:
            message_data['tts'] = True
        
        if (avatar_url is not None):
            if __debug__:
                if not isinstance(avatar_url, str):
                    raise AssertionError(
                        f'`avatar_url` can be `None`, `str`, got {avatar_url.__class__.__name__}; {avatar_url!r}.'
                    )
            
            message_data['avatar_url'] = avatar_url
        
        if (name is not None):
            if __debug__:
                if not isinstance(name, str):
                    raise AssertionError(
                        f'`name` cane be given either as `None`, `str instance, got '
                        f'{name.__class__.__name__}; {name!r}.'
                    )
                
                name_length = len(name)
                if name_length > 32:
                    raise AssertionError(
                        f'`name` length can be in range [1:32], got {name_length}; {name!r}.'
                    )
            
            if name:
                message_data['username'] = name
        
        message_data = add_file_to_message_data(message_data, file, contains_content, False)
        if message_data is None:
            return
        
        query_parameters = None
        if wait:
            if query_parameters is None:
                query_parameters = {}
            
            query_parameters['wait'] = wait
        
        if thread_id:
            if query_parameters is None:
                query_parameters = {}
            
            query_parameters['thread_id'] = thread_id
        
        message_data = await self.http.webhook_message_create(webhook_id, webhook_token, message_data, query_parameters)
        
        if not wait:
            return
        
        # Use goto
        while True:
            if (webhook is not None):
                channel = webhook.channel
                if (channel is not None):
                    break
            
            channel_id = int(message_data['channel_id'])
            channel = create_partial_channel_from_id(channel_id, ChannelType.guild_text, 0)
            break
        
        return channel._create_new_message(message_data)
    
    
    async def webhook_message_edit(
        self, webhook, message, content=..., *, embed = ..., file=..., allowed_mentions = ..., components = ...
    ):
        """
        Edits the message sent by the given webhook. The message's author must be the webhook itself.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook who created the message.
        
        message : ``Message``, `int`
            The webhook's message to edit.
        
        content : `None`, `str`, ``EmbedBase``, `Any`, Optional
            The new content of the message.
            
            If given as `str` then the message's content will be edited with it. If given as any non ``EmbedBase``
            instance, then it will be cased to string first.
            
            If given as ``EmbedBase``, then the message's embeds will be edited with it.
        
        embed : `None`, ``EmbedBase``, `list` of ``EmbedBase``, Optional (Keyword only)
            The new embedded content of the message. By passing it as `None`, you can remove the old.
            
            > If `embed` and `content` parameters are both given as  ``EmbedBase``, then `AssertionError` is
            raised.
        
        file : `None`, `Any`, Optional (Keyword only)
            A file or files to send. Check ``create_file_form`` for details.
        
        allowed_mentions : `None`,  `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions``
            for details.
        
        components : `None`, ``Component``, (`tuple`, `list`) of (``Component``, (`tuple`, `list`) of
                ``Component``), Optional (Keyword only)
            Components attached to the message.
        
        Raises
        ------
        TypeError
            - If `webhook` was not given neither as ``Webhook`` neither as a `tuple` (`int`, `str`).
            - If `allowed_mentions` contains an element of invalid type.
            - If `embed` was not given neither as ``EmbedBase`` nor as `list`, `tuple` of ``EmbedBase``-s.
            - `content` parameter was given as ``EmbedBase``, meanwhile `embed` parameter was given as well.
            - `message` was given as `None`. Make sure to use ``Client.webhook_message_create`` with `wait=True` and by
                giving any content to it as well.
            - `message` was not given neither as ``Message``, `int`.
            - If `components` type is incorrect.
        ValueError
            - If `allowed_mentions`'s elements' type is correct, but one of their value is invalid.
            - If more than `10` file would be sent.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `message` was detectably not sent by the `webhook`.
            - If `embed` contains a non ``EmbedBase`` element.
            - If both `content` and `embed` fields are embeds.
        
        See Also
        --------
        - ``.message_edit`` : Edit your own messages.
        - ``.webhook_message_create`` : Create a message with a webhook.
        - ``.webhook_message_delete`` : Delete a message created by a webhook.
        - ``.webhook_message_get`` : Get a message created by a webhook.
        
        Notes
        -----
        Embed messages ignore suppression with their endpoint, not like ``.message_edit`` endpoint.
        
        Editing the message with empty string is broken.
        """
        webhook_id, webhook_token = get_webhook_id_and_token(webhook)
        
        # Detect message id
        # 1.: Message
        # 2.: int (str)
        # 4.: None -> raise
        # 5.: raise
        
        if isinstance(message, Message):
            if __debug__:
                if message.author.id != webhook_id:
                    raise AssertionError(
                        f'The message was not sent by the webhook, got {message!r}.'
                    )
                
            message_id = message.id
        else:
            message_id = maybe_snowflake(message)
            if (message_id is not None):
                pass
            
            elif message is None:
                raise TypeError(
                    f'`message` was given as `None`. Make sure to use '
                    f'`{self.__class__.__name__}.webhook_message_create` with giving content and by passing '
                    f'`wait` parameter as `True` as well.'
                )
            
            else:
                raise TypeError(
                    f'`message` can be `{Message.__name__}`, int`, got '
                    f'{message.__class__.__name__}; {message!r}.'
                )
        
        content, embed = validate_content_and_embed(content, embed, True)
        
        components = get_components_data(components, True)
        
        # Build payload
        message_data = {}
        
        # Discord docs say, content can be nullable, but nullable content is just ignored.
        if (content is not ...):
            message_data['content'] = content
        
        if (embed is not ...):
            if (embed is not None):
                embed = [embed.to_data() for embed in embed]
            
            message_data['embeds'] = embed
        
        if (allowed_mentions is not ...):
            message_data['allowed_mentions'] = parse_allowed_mentions(allowed_mentions)
        
        if (components is not ...):
            message_data['components'] = components
        
        message_data = add_file_to_message_data(message_data, file, True, True)
        
        # We receive the new message data, but we do not update the message, so dispatch events can get the difference.
        await self.http.webhook_message_edit(webhook_id, webhook_token, message_id, message_data)
    
    
    async def webhook_message_delete(self, webhook, message):
        """
        Deletes the message sent by the webhook.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook who created the message.
        message : ``Message``, `int`
            The webhook's message to delete.
        
        Raises
        ------
        TypeError
            - If `message` was not given neither as ``Message``, `int`.
            - If `webhook` was not given neither as ``Webhook`` neither as a `tuple` (`int`, `str`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `message` was detectably not sent by the `webhook`.
        
        See Also
        --------
        - ``.message_delete`` : Delete a message.
        - ``.webhook_message_create`` : Create a message with a webhook.
        - ``.webhook_message_edit`` : Edit a message created by a webhook.
        - ``.webhook_message_get`` : Get a message created by a webhook.
        """
        webhook_id, webhook_token = get_webhook_id_and_token(webhook)
        
        # Detect message id
        # 1.: Message
        # 2.: int
        # 4.: None -> raise
        # 5.: raise
        
        if isinstance(message, Message):
            if __debug__:
                if message.author.id != webhook_id:
                    raise AssertionError(
                        f'The message was not sent by the webhook, got {message!r}.'
                    )
            
            message_id = message.id
        
        else:
            message_id = maybe_snowflake(message)
            if (message_id is not None):
                pass
            
            elif message is None:
                raise TypeError(
                    f'`message` was given as `None`. Make sure to use '
                    f'`{self.__class__.__name__}.webhook_message_create` with giving content and by passing '
                    f'`wait` parameter as `True` as well.'
                )
            
            else:
                raise TypeError(
                    f'`message` can be `{Message.__name__}`, `int`, got '
                    f'{message.__class__.__name__}; {message!r}.'
                )
        
        await self.http.webhook_message_delete(webhook_id, webhook_token, message_id)
    
    
    async def webhook_message_get(self, webhook, message_id):
        """
        Gets a previously sent message with the webhook.
        
        This method is a coroutine.
        
        Parameters
        ----------
        webhook : ``Webhook``, `tuple` (`int`, `str`)
            The webhook who created the message.
        message_id : `int`
            The webhook's message's identifier to get.
        
        Returns
        -------
        message : ``Message``
        
        Raises
        ------
        TypeError
            - If `webhook` was not given neither as ``Webhook`` neither as a `tuple` (`int`, `str`).
            - If `message_id` was not given as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        - ``.message_get`` : Get a message.
        - ``.webhook_message_create`` : Create a message with a webhook.
        - ``.webhook_message_edit`` : Edit a message created by a webhook.
        - ``.webhook_message_delete`` : Delete a message created by a webhook.
        """
        webhook_id, webhook_token = get_webhook_id_and_token(webhook)
        
        message_id_value = maybe_snowflake(message_id)
        if message_id_value is None:
            raise TypeError(
                f'`message_id` can be `int`, got {message_id.__class__.__name__}; {message_id!r}.'
            )
        
        message_data = await self.http.webhook_message_get(webhook_id, webhook_token, message_id_value)
        return Message(message_data)



__all__ = ('ChannelMetadataPrivateBase',)

from scarletio import copy_docs

from .fields import parse_users, put_users_into, validate_users

from .base import ChannelMetadataBase


class ChannelMetadataPrivateBase(ChannelMetadataBase):
    """
    Channel metadata for private channels.
    
    Attributes
    ----------
    users : `list` of ``ClientUserBase``
        The users in the channel.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('users',)
    
    @copy_docs(ChannelMetadataBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataBase.__hash__(self)
        
        # users
        users = self.users
        hash_value ^= len(users) << 13
        
        for user in users:
            hash_value ^= hash(user)
        
        return hash_value
    
    
    @classmethod
    @copy_docs(ChannelMetadataBase.from_data)
    def from_data(cls, data):
        self = super(ChannelMetadataPrivateBase, cls).from_data(data)
        
        # users
        self.users = parse_users(data)
        
        return self
    
    
    @copy_docs(ChannelMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if self.users != other.users:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataBase._get_users)
    def _get_users(self, channel_entity):
        return self.users
    
    
    @copy_docs(ChannelMetadataBase._iter_users)
    def _iter_users(self, channel_entity):
        yield from self.users
    
    
    @copy_docs(ChannelMetadataBase.name)
    def name(self):
        users = self.users
        if len(users) == 2:
            name = f'Direct Message {users[0].full_name} with {users[1].full_name}'
        else:
            name = f'Direct Message (partial)'
        return name
    
    
    @classmethod
    @copy_docs(ChannelMetadataBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataPrivateBase, cls)._create_empty()
        
        self.users = []
        
        return self
    

    @copy_docs(ChannelMetadataBase._set_attributes_from_keyword_parameters)
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        ChannelMetadataBase._set_attributes_from_keyword_parameters(self, keyword_parameters)
        
        # users
        try:
            users = keyword_parameters.pop('users')
        except KeyError:
            pass
        else:
            self.users = validate_users(users)
    
    
    @copy_docs(ChannelMetadataBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # users
        if include_internals:
            put_users_into(self.users, data, defaults)
        
        return data



__all__ = ()

from ...field_parsers import (
    bool_parser_factory, entity_id_parser_factory, force_string_parser_factory, int_parser_factory,
    nullable_string_parser_factory
)
from ...field_putters import (
    bool_optional_putter_factory, entity_id_putter_factory, force_string_putter_factory, int_putter_factory,
    nullable_string_putter_factory, url_optional_putter_factory
)
from ...field_validators import (
    bool_validator_factory, entity_id_validator_factory, force_string_validator_factory,
    int_conditional_validator_factory, nullable_string_validator_factory, url_optional_validator_factory
)

from .constants import DESCRIPTION_LENGTH_MAX

# content_type

parse_content_type = nullable_string_parser_factory('content_type')
put_content_type_into = nullable_string_putter_factory('content_type')
validate_content_type = nullable_string_validator_factory('content_type', 0, 1024)

# description

parse_description = nullable_string_parser_factory('description')
put_description_into = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', 0, DESCRIPTION_LENGTH_MAX)

# height

parse_height = int_parser_factory('height', 0)
put_height_into = int_putter_factory('height')
validate_height = int_conditional_validator_factory(
    'height',
    lambda height : height >= 0,
    '>= 0',
)

# id

parse_id = entity_id_parser_factory('id')
put_id_into = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('id')

# name

parse_name = force_string_parser_factory('filename')
put_name_into = force_string_putter_factory('filename')
validate_name = force_string_validator_factory('name', 0, 1024)

# proxy_url

parse_proxy_url = nullable_string_parser_factory('proxy_url')
put_proxy_url_into = url_optional_putter_factory('proxy_url')
validate_proxy_url = url_optional_validator_factory('proxy_url')

# size

parse_size = int_parser_factory('size', 0)
put_size_into = int_putter_factory('size')
validate_size = int_conditional_validator_factory(
    'size',
    lambda size : size >= 0,
    '>= 0',
)

# temporary

parse_temporary = bool_parser_factory('ephemeral', False)
put_temporary_into = bool_optional_putter_factory('ephemeral', False)
validate_temporary = bool_validator_factory('temporary')

# url

parse_url = force_string_parser_factory('url')
put_url_into = url_optional_putter_factory('url')
validate_url = url_optional_validator_factory('url')

# width

parse_width = int_parser_factory('width', 0)
put_width_into = int_putter_factory('width')
validate_width = int_conditional_validator_factory(
    'width',
    lambda width : width >= 0,
    '>= 0',
)



__all__ = ()

from .preinstanced import Locale


def serializable_localized_item_validator(item, parameter_name):
    """
    Json serializable localization dictionary item validator.
    
    Parameters
    ---------
    key : `tuple` ((``Locale``, `str`), `str`)
        An item representing a `locale` - `str` pair.
    parameter_name : `str`
        The parameter's name to raise exception with.
    
    Returns
    -------
    validated_key : `str`
    
    Raises
    ------
    TypeError
        If `key`'s type is incorrect.
    ValueError
        If `key`'s is an empty string.
    """
    key, value = item
    if isinstance(key, Locale):
        validated_key = key.value
    
    elif isinstance(key, str):
        if not key:
            raise ValueError(
                f'`{parameter_name}` keys cannot be empty strings, got item={item!r}.'
            )
        
        validated_key = key
    
    else:
        raise TypeError(
            f'`{parameter_name}` keys can be `{Locale.__name__}`, `str`, got {value.__class__.__name__}; {value!r}; '
            f'item={item!r}.'
        )
    
    if isinstance(value, str):
        if not value:
            raise ValueError(
                f'`{parameter_name}` values cannot be empty strings, got item={item!r}.'
            )
        
        validated_value = value
    
    else:
        raise TypeError(
            f'`{parameter_name}` values can be `str`, got {value.__class__.__name__}; {value!r}; item={item!r}.'
        )
    
    
    return validated_key, validated_value



def localized_dictionary_item_validator(item, parameter_name):
    """
    Localization dictionary item validator.
    
    Parameters
    ---------
    item : `tuple` ((``Locale``, `str`), `str`)
        An item representing a `locale` - `str` pair.
    parameter_name : `str`
        The parameter's name to raise exception with.
    
    Returns
    -------
    validated_key : `str`
    
    Raises
    ------
    TypeError
        If `key`'s type is incorrect.
    ValueError
        If `key`'s is an empty string.
    """
    key, value = item
    if isinstance(key, Locale):
        validated_key = key
    
    elif isinstance(key, str):
        if not key:
            raise ValueError(
                f'`{parameter_name}` keys cannot be empty strings, got item={item!r}.'
            )
        
        validated_key = Locale.get(key)
    
    else:
        raise TypeError(
            f'`{parameter_name}` keys can be `{Locale.__name__}`, `str`, got {value.__class__.__name__}; {value!r}; '
            f'item={item!r}.'
        )
    
    if isinstance(value, str):
        if not value:
            raise ValueError(
                f'`{parameter_name}` values cannot be empty strings, got item={item!r}.'
            )
        
        validated_value = value
    
    else:
        raise TypeError(
            f'`{parameter_name}` values can be `str`, got {value.__class__.__name__}; {value!r}; item={item!r}.'
        )
    
    
    return validated_key, validated_value


def localized_dictionary_builder(dictionary, parameter_name):
    """
    Parameters
    ----------
    dictionary : `None`, `dict` of ((``Locale``, `str`), `str`) items,
            (`set`, `tuple`, `list`) of `tuple` ((``Locale``, `str`), `str`)
        The value to convert to localized dictionary.
    parameter_name : `str`
        The parameter's name to raise exception with.
    
    Returns
    -------
    validated_dictionary : `None`, `dict` of (``Locale``, `Any`) items
    
    Raises
    ------
    TypeError
        - If `dictionary`'s or any of it's item's type is incorrect.
    ValueError
        - Empty key or value.
        - Incorrect item length.
    """
    return _dictionary_builder(dictionary, parameter_name, localized_dictionary_item_validator)


def serializable_localized_dictionary_builder(dictionary, parameter_name):
    """
    Builds a json serializable localized dictionary from the given dictionary.
    
    Parameters
    ----------
    dictionary : `None`, `dict` of ((``Locale``, `str`), `str`) items,
            (`set`, `tuple`, `list`) of `tuple` ((``Locale``, `str`), `str`)
        The value to convert to json serializable localized dictionary.
    parameter_name : `str`
        The parameter's name to raise exception with.
    
    Returns
    -------
    validated_dictionary : `None`, `dict` of (`str`, `Any`) items
    
    Raises
    ------
    TypeError
        - If `dictionary`'s or any of it's item's type is incorrect.
    ValueError
        - Empty key or value.
        - Incorrect item length.
    """
    return _dictionary_builder(dictionary, parameter_name, serializable_localized_item_validator)


def _dictionary_builder(dictionary, parameter_name, item_validator):
    """
    Builds a dictionary item validated by the given validator from the given one.
    
    Parameters
    ----------
    dictionary : `None`, `dict`, (`set`, `tuple`, `list`) of `tuple[2]`
        The value to convert.
    parameter_name : `str`
        The parameter's name to raise exception with.
    item_validator : `FunctionType`
        Item validator to pass every every item to.
    
    Returns
    -------
    validated_dictionary : `None`, `dict` of (`str`, `Any`) items
    
    Raises
    ------
    TypeError
        - If `dictionary`'s or any of it's item's type is incorrect.
    ValueError
        - Item structure incorrect.
    """
    if dictionary is None:
        validated_dictionary = None
    
    elif isinstance(dictionary, dict):
        validated_dictionary = {}
        
        for item in dictionary.items():
            key, value = item_validator(item, parameter_name)
            validated_dictionary[key] = value
        
        if not validated_dictionary:
            validated_dictionary = None
    
    elif isinstance(dictionary, (tuple, list, set)):
        validated_dictionary = {}
        for item in dictionary:
            if not isinstance(item, tuple):
                raise TypeError(
                    f'`{parameter_name}` items can be `tuple`-s, got {item.__class__.__name__}; {item!r}; '
                    f'{parameter_name}={dictionary!r}.'
                )
            
            
            item_length = len(item)
            if item_length != 2:
                raise ValueError(
                    f'`{parameter_name}` items can be `tuple`-s of length `2`, got {item!r}; length={item_length!r}; '
                    f'{parameter_name}={dictionary!r}.'
                )
            
            key, value = item_validator(item, parameter_name)
            validated_dictionary[key] = value
        
        if not validated_dictionary:
            validated_dictionary = None
            
    else:
        raise TypeError(
            f'`{parameter_name}` can be `None`, `dict`, `set`, `tuple`, `list`, '
            f'got {dictionary.__class__.__name__}; {dictionary!r}.'
        )
    
    return validated_dictionary


def get_localized_length(value, value_localizations):
    """
    Gets the length of the given value and of it's localized values.
    
    If a value has localizations, we only count the longest towards the overall length.
    
    Parameters
    ----------
    value : `None, `str`
        The default value.
    value_localizations : `None`, `dict` of (``Locale``, `str`) items
        Localizations of the value.
    
    Returns
    -------
    length : `int`
    """
    if value is None:
        length = 0
    else:
        length = len(value)
    
    if (value_localizations is not None):
        for value in value_localizations.values():
            length = max(len(value), length)
    
    return length



__all__ = ()

from ....bases import Icon
from ....color import Color
from ....permission import Permission
from ....utils import timestamp_to_datetime

from ..audit_log_change import AuditLogChange


def _convert_preinstanced(name, data, preinstanced_type):
    before = data.get('old_value', None)
    if (before is not None):
        before = preinstanced_type.get(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = preinstanced_type.get(after)
    
    return AuditLogChange(name, before, after)


def convert_color(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = Color(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = Color(after)
    
    return AuditLogChange('color', before, after)


def convert_deprecated(name, data):
    return None


def convert_icon(name, data):
    if name == 'splash_hash':
        name = 'invite_splash'
    elif name.endswith('hash'):
        name = name[:-5]
    
    before = Icon.from_base16_hash(data.get('old_value', None))
    
    after = Icon.from_base16_hash(data.get('new_value', None))
    
    return AuditLogChange(name, before, after)


def convert_nothing(name, data):
    before = data.get('old_value', None)
    after = data.get('new_value', None)
    return AuditLogChange(name, before, after)


def convert_permission(name, data):
    if name.endswith('_new'):
        name = name[:-4]
    
    before = data.get('old_value', None)
    if (before is not None):
        before = Permission(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = Permission(after)
    
    return AuditLogChange(name, before, after)


def convert_snowflake(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        before = int(before)
    
    after = data.get('new_value', None)
    if (after is not None):
        after = int(after)
    
    return AuditLogChange(name, before, after)


def convert_timestamp(name, data):
    before = data.get('old_value', None)
    if (before is None):
        before = None
    else:
        before = timestamp_to_datetime(before)
    
    after = data.get('new_value', None)
    if (after is None):
        after = None
    else:
        after = timestamp_to_datetime(after)
    
    return AuditLogChange(name, before, after)


def convert_snowflake_array(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        if before:
            before = tuple(int(sub_value) for sub_value in before)
        else:
            before = None
    
    after = data.get('new_value', None)
    if (after is not None):
        if after:
            after = tuple(int(sub_value) for sub_value in after)
        else:
            after = None
    
    return AuditLogChange(name, before, after)



__all__ = ('PluginError', )

from scarletio import CauseGroup


class PluginError(Exception):
    """
    An exception raised by the ``PluginLoader``, if loading, reloading or unloading an plugin fails with any
    reason.
    
    Attributes
    ----------
    _message : `None`, `str`
        The error's message.
    """
    
    def __init__(self, message = None, *, cause = None):
        """
        Creates a new plugin error.
        
        Parameters
        ----------
        message : `None`, `str` = `None`, Optional
            The error's message.
        
        cause : `None`, `BaseException` = `None`, Optional (Keyword only)
            Exception cause to apply manually.
        """
        self._message = message
        
        if message is None:
            Exception.__init__(self)
        else:
            Exception.__init__(self, message)
        
        if (cause is not None):
            self.__cause__ = cause
    
    
    @property
    def message(self):
        """
        Returns the plugin error's message.
        
        If the plugin error contains more messages connects them.
        
        Returns
        -------
        message : ``Message``
        """
        return '\n\n'.join(self.messages)
    
    
    @property
    def messages(self):
        """
        Returns a list containing the exception error's messages.
        
        Returns
        -------
        messages : `list` of `str`
        """
        messages = []
        
        message = self._message
        if (message is not None):
            messages.append(message)
        
        cause = self.__cause__
        if (cause is not None):
            if isinstance(cause, CauseGroup):
                for cause in cause:
                    messages.append(repr(cause))
            
            else:
                messages.append(repr(cause))
        
        return message
    
    
    def __len__(self):
        """Returns the amount of messages, what the plugin error contains."""
        length = 0
        
        if (self._message is not None):
            length += 1

        cause = self.__cause__
        if (cause is not None):
            if isinstance(cause, CauseGroup):
                length += len(cause)
            
            else:
                length += 1
        
        return length
    
    
    def __repr__(self):
        """Returns the exception error's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        message = self._message
        if (message is not None):
            repr_parts.append(' message=')
            repr_parts.append(repr(message))
        
        cause = self.__cause__
        if (cause is not None):
            if isinstance(cause, CauseGroup):
                cause_count = len(cause)
            
            else:
                cause_count = 1
            
            repr_parts.append(' with ')
            repr_parts.append(repr(cause_count))
            repr_parts.append(' cause')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    __str__ = __repr__


class DoNotLoadPlugin(BaseException):
    """
    Raised to stop an plugin loaded without error.
    """



__all__ = ('ChannelMetadataGuildStore', )

from scarletio import copy_docs

from ...permission import Permission
from ...permission.permission import PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_NONE, PERMISSION_TEXT_AND_VOICE_DENY

from .fields import parse_nsfw, put_nsfw_into, validate_nsfw

from .guild_main_base import ChannelMetadataGuildMainBase


class ChannelMetadataGuildStore(ChannelMetadataGuildMainBase):
    """
    Guild store channel metadata.
    
    Attributes
    ----------
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    name : `str`
        The channel's name.
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    nsfw : `bool`
        Whether the channel is marked as non safe for work.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('nsfw',)
    
    @copy_docs(ChannelMetadataGuildMainBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataGuildMainBase.__hash__(self)
        
        # nsfw
        hash_value ^= self.nsfw << 28
        
        return hash_value
    
    
    @copy_docs(ChannelMetadataGuildMainBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataGuildMainBase._is_equal_same_type(self, other):
            return False
        
        # nsfw
        if self.nsfw != other.nsfw:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_display_name)
    def _get_display_name(self):
        return self.name.lower()
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildMainBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildStore, cls)._create_empty()
        
        self.nsfw = False
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildMainBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildMainBase._update_attributes(self, data)
        
        # nsfw
        self.nsfw = parse_nsfw(data)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildMainBase._difference_update_attributes(self, data)
        
        # nsfw
        nsfw = parse_nsfw(data)
        if self.nsfw != nsfw:
            old_attributes['nsfw'] = self.nsfw
            self.nsfw = nsfw
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        result = self._get_base_permissions_for(channel_entity, user)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # store channels do not have text and voice related permissions
        result &= PERMISSION_TEXT_AND_VOICE_DENY
        
        return Permission(result)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        result = self._get_base_permissions_for_roles(channel_entity, roles)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # store channels do not have text and voice related permissions
        result &= PERMISSION_TEXT_AND_VOICE_DENY
        return Permission(result)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._set_attributes_from_keyword_parameters)
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        ChannelMetadataGuildMainBase._set_attributes_from_keyword_parameters(self, keyword_parameters)
        
        # nsfw
        try:
            nsfw = keyword_parameters.pop('nsfw')
        except KeyError:
            pass
        else:
            self.nsfw = validate_nsfw(nsfw)
    
    
    @copy_docs(ChannelMetadataGuildMainBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataGuildMainBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        put_nsfw_into(self.nsfw, data, defaults)
        
        return data



__all__ = ()

from scarletio import RichAttributeErrorBaseType

from .render_constants import (
    PARAMETER_COLUMN_SEPARATOR, PARAMETER_DEFAULT_PREFIX, PARAMETER_MODIFIER_SEPARATOR, PARAMETER_NOTE_ADJUSTMENT,
    PARAMETER_NOTE_SIGN_MULTIPLE_KEYWORD, PARAMETER_NOTE_SIGN_MULTIPLE_POSITIONAL, PARAMETER_NOTE_SIGN_REQUIRED,
    PARAMETER_TYPE_IDENTIFIER_TO_REPRESENTATION
)


class ParameterRenderer(RichAttributeErrorBaseType):
    """
    Helper class for rendering parameter representation.
    
    Attributes
    ----------
    _default_representation : `None`, `str`
        The parameter's default value's representation.
    _default_representation_set : `bool`
        Whether ``._default_representation`` was already set.
    command_parameter : ``CommandParameter``
        The represented parameter.
    """
    __slots__ = ('_default_representation', '_default_representation_set', 'command_parameter',)
    
    def __new__(cls, command_parameter):
        """
        Creates a new parameter renderer instance.
        
        Parameters
        ----------
        command_parameter : ``CommandParameter``
            The represented parameter.
        """
        self = object.__new__(cls)
        self._default_representation = None
        self._default_representation_set = False
        self.command_parameter = command_parameter
        return self
    
    
    def get_name_length(self):
        """
        Gets the parameter's name's length.
        
        Returns
        -------
        name_length : `int`
        """
        command_parameter = self.command_parameter
        
        name_length = len(command_parameter.display_name)
        
        if command_parameter.is_modifier():
            name_length += 2
        else:
            parameter = command_parameter.parameter
            if parameter.is_keyword_only():
                name_length += 1
        
        return name_length
    
    
    def get_reverted_name_length(self):
        """
        Gets the reverted name of the option.
        
        Returns
        -------
        reverted_name_length : `int`
        """
        return self.get_name_length() + 3
    
    
    def get_name(self):
        """
        Returns the parameter's name.
        
        Returns
        -------
        name : `str`
        """
        command_parameter = self.command_parameter
        
        name = command_parameter.display_name
        
        if command_parameter.is_modifier():
            name = f'--{name}'
        else:
            parameter = command_parameter.parameter
            if parameter.is_keyword_only():
                name = f'-{name}'
        
        return name
    
    
    def get_reverted_name(self):
        """
        gets the reverted name of the parameter.
        
        Returns
        -------
        reverted_name : `str`
        """
        return f'--no-{self.command_parameter.display_name}'
    
    
    def get_type_name_length(self):
        """
        Returns the length of the parameter's type.
        
        Returns
        -------
        type_name_length : `int`
        """
        return len(self.get_type_name())
    
    
    def get_type_name(self):
        """
        Returns the type name of the parameter.
        
        Returns
        -------
        type_name : `str`
        """
        return PARAMETER_TYPE_IDENTIFIER_TO_REPRESENTATION[self.command_parameter.expected_type_identifier]
    
    
    def get_default_length(self):
        """
        Gets the parameter's default value's length.
        
        Returns
        -------
        default_length : `int`
        """
        default = self.get_default()
        if default is None:
            default_length = 0
        else:
            default_length = len(default)
        
        return default_length
    
    
    def get_default(self):
        """
        Gets the parameter's default value's representation.
        
        Returns
        -------
        default : `None`, `str`
        """
        if self._default_representation_set:
            return self._default_representation
        
        command_parameter = self.command_parameter
        parameter = command_parameter.parameter
        
        if parameter.has_default:
            default = parameter.default
            if command_parameter.is_modifier():
                default_representation = command_parameter.display_name
                if not default:
                    default_representation = f'no-{default_representation}'
            else:
                if default is ...:
                    default_representation = None
                else:
                    default_representation = repr(default)
        else:
            default_representation = None
        
        self._default_representation_set = True
        self._default_representation = default_representation
        return default_representation
    
    
    def is_required(self):
        """
        Returns whether the parameter is required.
        
        Returns
        -------
        is_required : `bool`
        """
        return self.command_parameter.is_required()
    
    
    def is_positional(self):
        """
        Returns whether the parameter is positional.
        
        Returns
        -------
        is_positional : `bool`
        """
        return self.command_parameter.parameter.is_positional()
    
    
    def is_args(self):
        """
        Returns whether self represents an `*args` parameter.
        
        Returns
        -------
        is_args : `bool`
        """
        return self.command_parameter.parameter.is_args()
    
    
    def is_kwargs(self):
        """
        Returns whether self represents an `**kwargs` parameter.
        
        Returns
        -------
        is_kwargs : `bool`
        """
        return self.command_parameter.parameter.is_kwargs()
    
    
    def is_modifier(self):
        """
        Returns whether the command parameter is a modifier one.
        
        Returns
        -------
        is_modifier : `bool`
        """
        return self.command_parameter.is_modifier()
    
    
    def get_note_sign(self):
        """
        Returns the note sign's length.
        
        Returns
        -------
        note_sign : `None`, `str`
        """
        if self.is_required():
            return PARAMETER_NOTE_SIGN_REQUIRED
        
        if self.is_args():
            return PARAMETER_NOTE_SIGN_MULTIPLE_POSITIONAL
        
        if self.is_kwargs():
            return PARAMETER_NOTE_SIGN_MULTIPLE_KEYWORD
        
        return None
        
    
    def get_note_sign_length(self):
        """
        Returns the note sign's length.
        
        Returns
        -------
        note_sign_length : `int`
        """
        note_sign = self.get_note_sign()
        if note_sign is None:
            note_sign_length = 0
        else:
            note_sign_length = len(note_sign)
        
        return note_sign_length
    
    
    def get_generic_lengths(self):
        """
        Returns the parameter's lengths.
        
        Returns
        -------
        lengths : `tuple` of `int`
        """
        return (
            self.get_note_sign_length(),
            self.get_name_length(),
            self.get_type_name_length(),
            self.get_default_length(),
        )
    
    
    def get_modifier_lengths(self):
        """
        Gets the parameter's modifier form lengths.
        
        Returns
        -------
        lengths : `tuple` of `int`
        """
        return (
            self.get_name_length(),
            self.get_reverted_name_length(),
            self.get_default_length(),
        )
    
    
    def render_generic_into_with_lengths(self, into, lengths):
        """
        Renders the parameter into the given lost with the given lengths.
        
        > This rendering method is for generic parameters.
        
        Parameters
        ----------
        into : `list` of `str`
            The list to render the string parts into.
        lengths : `tuple` of `int`
            Lengths representing each column's width.
        
        Returns
        -------
        into : `list` of `str`
        """
        note_length, name_length, type_name_length, default_length = lengths
        
        if note_length:
            into = _render_part_into(into, self.get_note_sign(), PARAMETER_NOTE_ADJUSTMENT)
        
        into = _render_part_into(into, self.get_name(), name_length)
        into.append(PARAMETER_COLUMN_SEPARATOR)
        into = _render_part_into(into, self.get_type_name(), type_name_length)
        
        if default_length:
            into.append(PARAMETER_COLUMN_SEPARATOR)
            default = self.get_default()
            into = _render_part_into(
                into, None if (default is None) else PARAMETER_DEFAULT_PREFIX, len(PARAMETER_DEFAULT_PREFIX)
            )
            into = _render_part_into(into, default, default_length)
        
        return into
    
    
    def render_modifier_into_with_lengths(self, into, lengths):
        """
        Renders the parameter into the given lost with the given lengths.
        
        > This rendering method is for modifier parameters.
        
        Parameters
        ----------
        into : `list` of `str`
            The list to render the string parts into.
        lengths : `tuple` of `int`
            Lengths representing each column's width.
        
        Returns
        -------
        into : `list` of `str`
        """
        name_length, reverted_name_length, default_length = lengths
        
        into = _render_part_into(into, self.get_name(), name_length)
        into.append(PARAMETER_MODIFIER_SEPARATOR)
        into = _render_part_into(into, self.get_reverted_name(), reverted_name_length)
        
        if default_length:
            into.append(PARAMETER_COLUMN_SEPARATOR)
            default = self.get_default()
            into = _render_part_into(
                into, None if (default is None) else PARAMETER_DEFAULT_PREFIX, len(PARAMETER_DEFAULT_PREFIX)
            )
            into = _render_part_into(into, default, default_length)
        
        return into


def get_render_generic_line_length(lengths):
    """
    Gets the output line length for the given `lengths` variable.
    
    > This utility function is for generic parameters.
    
    Parameters
    ----------
    lengths : `None`, `tuple` of `int`
        Lengths representing each column's width.
    
    Returns
    -------
    length : `int`
        The cumulate of the line.
    """
    if (lengths is None):
        return 0
    
    length = 0
    
    note_length, name_length, type_name_length, default_length = lengths
    
    if note_length:
        length += PARAMETER_NOTE_ADJUSTMENT
    
    length += name_length
    length += len(PARAMETER_COLUMN_SEPARATOR)
    length += type_name_length
    
    if default_length:
        length += len(PARAMETER_COLUMN_SEPARATOR)
        length += len(PARAMETER_DEFAULT_PREFIX)
        length += default_length
    
    return length


def get_render_modifier_line_length(lengths):
    """
    Gets the output line length for the given `lengths` variable.
    
    > This utility function is for modifier parameters.
        
    Parameters
    ----------
    lengths : `None`, `tuple` of `int`
        Lengths representing each column's width.
    
    Returns
    -------
    length : `int`
        The cumulate of the line.
    """
    if (lengths is None):
        return 0
    
    length = 0
    
    name_length, reverted_name_length, default_length = lengths
    
    length += name_length
    length += len(PARAMETER_MODIFIER_SEPARATOR)
    length += reverted_name_length
    
    if default_length:
        length += len(PARAMETER_COLUMN_SEPARATOR)
        length += len(PARAMETER_DEFAULT_PREFIX)
        length += default_length
    
    return length


def merge_lengths(lengths_1, lengths_2):
    """
    Merges the given length tuples.
    
    
    Parameters
    ----------
    lengths_1 : `tuple` of `int`
        Length tuple 1.
    lengths_2 : `tuple` of `int`
        Length tuple 2.
    
    Returns
    -------
    length_tuple : `tuple` of `int`
    """
    return tuple(max(value_1, value_2) for value_1, value_2 in zip(lengths_1, lengths_2))


def _render_part_into(into, part, length):
    """
    Renders he given part into the given container on the specified length.
    
    Parameters
    ----------
    into : `list` of `str`
        The list of strings to render to.
    part : `None`, `str`
        The part to render.
    length : `int`
        The length to fill up.
    
    Returns
    -------
    into : `list` of `str`
    """
    adjust_length = length
    
    if (part is not None):
        into.append(part)
        adjust_length -= len(part)
    
    if adjust_length:
        into.append(' ' * adjust_length)
    
    return into


def get_lengths(parameter_renderers, getter):
    """
    Gets the lengths of the elements of the given list.
    
    Parameters
    ----------
    parameter_renderers : `None`, `list` of ``ParameterRenderer``
        Parameters to get their length of.
    getter : `FunctionType`
        Getter to get lengths of the the parameter renderers.
    
    Returns
    -------
    lengths : `None`, `tuple` of `str`
        Returns `None` if the `parameter_renderers` is empty.
    """
    if (parameter_renderers is None) or (not parameter_renderers):
        return None
    
    iterator = iter(parameter_renderers)
    
    lengths = getter(next(iterator))
    for parameter_renderer in iterator:
        lengths = merge_lengths(lengths, getter(parameter_renderer))
    
    return lengths



__all__ = ()

from scarletio import ScarletLock, RichAttributeErrorBaseType, copy_docs

from ...discord.core import KOKORO


class RateLimitContextBase(RichAttributeErrorBaseType):
    """
    Rate limit context used to handle static rate limits when communicating with top.gg.
    
    Attributes
    ----------
    acquired : `bool`
        Whether the lock is acquired.
    """
    __slots__ = ('acquired', )
    
    def __new__(cls):
        """
        Creates a new rate limit context instance.
        """
        self = object.__new__(cls)
        self.acquired = False
        return self
    
    
    async def __aenter__(self):
        """
        Enters the rate limit context, blocking till acquiring it.
        
        This method is a coroutine.
        """
        self.acquired = True
        return self


    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Releases the rate limit context.
        
        This method is a coroutine.
        """
        self.release()
        return False
        
    
    def __del__(self):
        """Releases the rate limit context if not yet released."""
        self.release()
    
    
    def release(self):
        """Releases the rate limit context."""
        self.acquired = False


class RateLimitContext(RateLimitContextBase):
    """
    Rate limit context used to handle static rate limits when communicating with top.gg.
    
    Attributes
    ----------
    acquired : `bool`
        Whether the lock is acquired.
    rate_limit_group : ``RateLimitGroup``
        The parent rate limit group.
    """
    __slots__ = ('rate_limit_group',)
    
    def __new__(cls, rate_limit_group):
        """
        Creates a new rate limit context instance.
        
        Parameters
        ----------
        rate_limit_group : ``RateLimitGroup``
            The parent rate limit group.
        """
        self = object.__new__(cls)
        self.rate_limit_group = rate_limit_group
        self.acquired = False
        return self
    
    
    @copy_docs(RateLimitContextBase.__aenter__)
    async def __aenter__(self):
        await self.rate_limit_group.lock.acquire()
        self.acquired = True
        return self
    
    
    @copy_docs(RateLimitContextBase.release)
    def release(self):
        if self.acquired:
            rate_limit_group = self.rate_limit_group
            KOKORO.call_later(rate_limit_group.reset_after, rate_limit_group.lock.release)
            self.acquired = False


class StackedRateLimitContext(RateLimitContextBase):
    """
    Rate limit context used to handle multiple static rate limits when communicating with top.gg.
    
    Attributes
    ----------
    acquired : `bool`
        Whether the lock is acquired.
    rate_limit_groups : `tuple` of ``RateLimitGroup``
        The parent rate limit groups.
    """
    __slots__ = ('rate_limit_groups',)
    
    def __new__(cls, rate_limit_groups):
        """
        Creates a new rate limit context instance.
        
        Parameters
        ----------
        rate_limit_groups : `tuple` of ``RateLimitGroup``
            The parent rate limit groups.
        """
        self = object.__new__(cls)
        self.rate_limit_groups = rate_limit_groups
        self.acquired = False
        return self
    
    
    @copy_docs(RateLimitContextBase.__aenter__)
    async def __aenter__(self):
        # Use linear acquiring.
        for rate_limit_group in self.rate_limit_groups:
            try:
                await rate_limit_group.lock.acquire()
            except:
                # If already acquired, release on cancellation
                for rate_limit_group_to_cancel in self.rate_limit_groups:
                    if rate_limit_group is rate_limit_group_to_cancel:
                        break
                    
                    rate_limit_group_to_cancel.lock.release()
                    continue
                
                raise
        
        self.acquired = True
        return self
    
    
    @copy_docs(RateLimitContextBase.release)
    def release(self):
        if self.acquired:
            for rate_limit_group in self.rate_limit_groups:
                KOKORO.call_later(rate_limit_group.reset_after, rate_limit_group.lock.release)
            
            self.acquired = False


class RateLimitHandlerBase(RichAttributeErrorBaseType):
    """
    Rate limit handler which can be entered.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new rate limit handler instance.
        """
        return object.__new__(cls)
    
    def ctx(self):
        """
        Enters the rate limit handler, allowing it to be used as an asynchronous context manager.
        
        Returns
        -------
        rate_limit_context : ``RateLimitContextBase``
        """
        return RateLimitContextBase()


class RateLimitHandler(RateLimitHandlerBase):
    """
    Rate limit handler which can be entered.
    
    Attributes
    ----------
    rate_limit_group : ``RateLimitGroup``
        The wrapped rate limit group.
    """
    __slots__ = ('rate_limit_group',)
    
    def __new__(cls, rate_limit_group):
        """
        Creates a new rate limit handler instance with the given rate limit group.
        
        Parameters
        ----------
        rate_limit_group : ``RateLimitGroup``
            The parent rate limit group.
        """
        self = object.__new__(cls)
        self.rate_limit_group = rate_limit_group
        return self
    
    
    @copy_docs(RateLimitHandlerBase.ctx)
    def ctx(self):
        return RateLimitContext(self.rate_limit_group)


class StackedRateLimitHandler(RateLimitHandlerBase):
    """
    Rate limit handler which can be entered.
    
    Attributes
    ----------
    rate_limit_groups : `tuple` of ``RateLimitGroup``
        The wrapped rate limit group.
    """
    __slots__ = ('rate_limit_groups',)
    
    def __new__(cls, *rate_limit_groups):
        """
        Creates a new rate limit handler instance with the given rate limit group.
        
        Parameters
        ----------
        *rate_limit_groups : ``RateLimitGroup``
            The parent rate limit groups.
        """
        self = object.__new__(cls)
        self.rate_limit_groups = rate_limit_groups
        return self
    
    
    @copy_docs(RateLimitHandlerBase.ctx)
    def ctx(self):
        return StackedRateLimitContext(self.rate_limit_groups)


class RateLimitGroup(RichAttributeErrorBaseType):
    """
    Static rate limit handler group implementation.
    
    Attributes
    ----------
    lock : ``ScarletLock``
        Lock used to block requests.
    reset_after : `int`
        Duration to release a rate limit after.
    """
    __slots__ = ('lock', 'reset_after', )
    
    def __new__(cls, size, reset_after):
        """
        Creates a new rate limit handler instance.
        
        Parameters
        ----------
        size : `int`
            Rate limit size.
        reset_after : `float`
            The time to reset rate limit after.
        """
        self = object.__new__(cls)
        self.reset_after = reset_after
        self.lock = ScarletLock(KOKORO, size)
        return self



__all__ = ('CommandBase',)

from functools import partial as partial_func

from scarletio import RichAttributeErrorBaseType
from .....discord.events.handling_helpers import create_event_from_class

from ...exceptions import _register_exception_handler, test_exception_handler


class CommandBase(RichAttributeErrorBaseType):
    """
    Base type for ``Slasher``'s commands.
    
    Attributes
    ----------
    _exception_handlers : `None`, `list` of `CoroutineFunction`
        Exception handlers added with ``.error`` to the interaction handler.
        
        Same as ``Slasher._exception_handlers``.
    
    _parent_reference : `None`, ``WeakReferer`` to ``SlashCommand``
        The parent slasher of the component command.

    name : `str`
        The command's name.
        
        Only used for debugging.
    
    
    Class Attributes
    ----------------
    COMMAND_COMMAND_NAME : `str`
        The command's name defining parameter's name.
    COMMAND_PARAMETER_NAMES : `tuple` of `str`
        All parameters names accepted by ``.__new__``
    COMMAND_NAME_NAME : `str`
        The command's "command" defining parameter's name.
    """
    __slots__ = ('name', '_exception_handlers', '_parent_reference',)

    COMMAND_NAME_NAME = 'name'
    COMMAND_COMMAND_NAME = 'command'
    
    COMMAND_PARAMETER_NAMES = (
        COMMAND_NAME_NAME,
        COMMAND_COMMAND_NAME,
        'allowed_mentions',
        'wait_for_acknowledgement',
        'show_for_invoking_user_only',
    )
    
    
    @classmethod
    def from_class(cls, klass):
        """
        Creates a new command instance from the given `klass`.
        
        Parameters
        ----------
        klass : `type`
            The class to create custom id based command from.
        
        Returns
        -------
        self : ``CommandBase``, ``Router``
        
        Raises
        ------
        TypeError
            If any attribute's type is incorrect.
        ValueError
            If any attribute's value is incorrect.
        """
        return create_event_from_class(
            cls, klass, cls.COMMAND_PARAMETER_NAMES, cls.COMMAND_NAME_NAME, cls.COMMAND_COMMAND_NAME
        )


    def __new__(cls, func, name = None, **keyword_parameters):
        """
        Creates a new custom_id based command instance.
        
        Subclasses should overwrite this method.
        
        Parameters
        ----------
        func : `None`, `async-callable`
            The function used as the command when using the respective slash command.
        name : `None`, `str` = `None`, Optional
            The name of the component command.

        Returns
        -------
        self : ``CommandBase``, ``Router``
        
        Raises
        ------
        NotImplementedError
        """
        raise NotImplementedError
    
    
    def _cursed_repr_builder(self):
        """
        Representation builder helper.
        
        This method is a generator.
        
        Examples
        --------
        ```
        for repr_parts in self._cursed_repr_builder():
            repr_parts.append(', oh no')
        
        return ''.join(repr_parts)
        ```
        """
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' name=')
        repr_parts.append(repr(self.name))
        
        yield repr_parts
        
        repr_parts.append('>')
    
    
    def __repr__(self):
        """Returns the command's representation."""
        for repr_parts in self._cursed_repr_builder():
            pass
        
        return ''.join(repr_parts)
    
    
    def __format__(self, code):
        """Formats the command in a format string."""
        if not code:
            return str(self)
        
        if code == 'm':
            return self.mention
        
        raise ValueError(
            f'Unknown format code {code!r} for {self.__class__.__name__}; {self!r}. '
            f'Available format codes: {""!r}, {"m"!r}.'
        )
    
    
    async def invoke(self, client, interaction_event):
        """
        Calls the command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        interaction_event : ``InteractionEvent``
            The received interaction event.
        """
        return
    
    
    def __hash__(self):
        """Returns the command's hash value."""
        hash_value = 0
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            hash_value ^= len(exception_handlers) << 4
            
            for exception_handler in exception_handlers:
                try:
                    exception_handler_hash_value = hash(exception_handler)
                except TypeError:
                    exception_handler_hash_value = object.__hash__(exception_handler)
                hash_value ^= exception_handler_hash_value
        
        # _parent_reference
        # Internal field
        
        # name
        hash_value ^= hash(self.name)
        
        return hash_value
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two types are equal.
        
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `type(self)`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # _exception_handlers
        if self._exception_handlers != other._exception_handlers:
            return False
        
        # _parent_reference
        # Internal field
        
        # name
        if self.name != other.name:
            return False
        
        return True
    
    
    def __eq__(self, other):
        """Returns whether the two commands are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def copy(self):
        """
        Copies the command.
        
        Returns
        -------
        new : `type<self>`
        """
        new = object.__new__(type(self))
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            exception_handlers = exception_handlers.copy()
        new._exception_handlers = exception_handlers
        
        # _parent_reference
        new._parent_reference = None
        
        # name
        new.name = self.name
        
        return new
    
    
    def error(self, exception_handler = None, *, first = False):
        """
        Registers an exception handler to the command.
        
        Parameters
        ----------
        exception_handler : `None`, `CoroutineFunction` = `None`, Optional
            Exception handler to register.
        first : `bool` = `False`, Optional (Keyword Only)
            Whether the exception handler should run first.
        
        Returns
        -------
        exception_handler / wrapper : `CoroutineFunction` / `functools.partial`
            If `exception_handler` is not given, returns a wrapper.
        """
        if exception_handler is None:
            return partial_func(_register_exception_handler, first)
        
        return self._register_exception_handler(exception_handler, first)
    
    
    def _register_exception_handler(self, exception_handler, first):
        """
        Registers an exception handler to the command.
        
        Parameters
        ----------
        exception_handler : `CoroutineFunction`
            Exception handler to register.
        first : `bool`
            Whether the exception handler should run first.
        
        Returns
        -------
        exception_handler : `CoroutineFunction`
        """
        test_exception_handler(exception_handler)
        
        exception_handlers = self._exception_handlers
        if exception_handlers is None:
            self._exception_handlers = exception_handlers = []
        
        if first:
            exception_handlers.insert(0, exception_handler)
        else:
            exception_handlers.append(exception_handler)
        
        return exception_handler
    
    # ---- Mention ----
    
    @property
    def mention(self):
        """
        Returns the command mention.
        Applicable for slash and context commands.
        
        Returns
        -------
        mention : `str`
        """
        return ''
    
    
    def mention_at(self, guild):
        """
        Mentions the command at the specified guild.
        Should be used when the command is added to multiple guilds.
        Applicable for slash and context commands.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to get the command's mention.
        
        Returns
        -------
        mention : `str`
        """
        return ''



__all__ = ()

import warnings

from .. import plugin_loader


def __getattr__(attribute_name):
    new_attribute_name = attribute_name.replace(
        'extension', 'plugin'
    ).replace(
        'EXTENSION', 'PLUGIN'
    ).replace(
        'Extension', 'Plugin'
    )
    
    warnings.warn(
        (
            f'`extension_loader.{attribute_name}` extension is deprecated and will be removed in 2022 December, '
            f'please use `plugin_loader.{new_attribute_name}` instead.'
        ),
        FutureWarning,
        stacklevel = 2,
    )
    
    return getattr(plugin_loader, new_attribute_name)



__all__ = (
    'ICON_TYPE_ANIMATED', 'ICON_TYPE_NONE', 'ICON_TYPE_STATIC', 'Icon', 'IconType', 'IconSlot', 'PreinstancedBase'
)

import reprlib, sys, warnings
from base64 import b64encode

from scarletio import DOCS_ENABLED, RichAttributeErrorBaseType, copy_docs, docs_property, include

from .place_holder import PlaceHolder
from .preinstanced import Preinstance as P, PreinstancedBase

get_image_media_type = include('get_image_media_type')


class IconDetailsBase(RichAttributeErrorBaseType):
    """
    Stores details of an icon.
    """
    __slots__ = ()
    
    allowed_postfixes = PlaceHolder(
        None,
        """
        Returns allowed postfixes.
        
        Returns
        -------
        allowed_postfixes : `None`, `frozenset` of `str`
        """
    )
    
    data = PlaceHolder(
        None,
        """
        Returns the icon's raw data.
        
        Returns
        -------
        data : `None`, `bytes-like`
        """
    )
    
    default_postfix = PlaceHolder(
        '',
        """
        Returns the icon's default postfix.
        
        Returns
        -------
        default_postfix : `str`
        """
    )
    
    prefix = PlaceHolder(
        '',
        """
        Returns the icon's prefix.
        
        Returns
        -------
        prefix : `str`
        """
    )
    
    media_type = PlaceHolder(
        '',
        """
        Returns the icon's media type.
        
        Returns
        -------
        media_type : `str`
        """
    )
    
    def __repr__(self):
        """Returns the icon details' representation."""
        return f'<{self.__class__.__name__}>'
        

class IconDetailsPreinstanced(IconDetailsBase):
    """
    Represents details about a preinstanced icon type.
    
    Attributes
    ----------
    allowed_postfixes : `None`, `frozenset` of `str`
        The allowed postfixes.
    default_postfix : `str`
        Default postfix used when building an url with the icon.
    prefix : `str`
        Prefix used when building an url with the icon.
    """
    __slots__ = ('allowed_postfixes', 'default_postfix', 'prefix')
    
    def __init__(self, allowed_postfixes, default_postfix, prefix):
        """
        Creates a new icon type detail for storing information for a preinstanced icon type.
        
        Parameters
        ----------
        allowed_postfixes : `None`, `frozenset` of `str`
            The allowed postfixes.
        default_postfix : `str`
            Default postfix used when building an url with the icon.
        prefix : `str`
            Prefix used when building an url with the icon.
        """
        self.allowed_postfixes = allowed_postfixes
        self.default_postfix = default_postfix
        self.prefix = prefix
    
    
    @copy_docs(IconDetailsBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' allowed_postfixes=')
        repr_parts.append(repr(self.allowed_postfixes))
        
        repr_parts.append(', default_postfix=')
        repr_parts.append(repr(self.default_postfix))
        
        repr_parts.append(', prefix=')
        repr_parts.append(repr(self.prefix))
        
        repr_parts.append('>')
        return ''.join(repr_parts)


class IconDetailsCustom(IconDetailsBase):
    """
    Represents details fo a custom icon type.
    
    Attributes
    ----------
    data : `None`, `bytes-like`
        Icon payload.
    media_type : `str`
        The icon's data's media type.
    """
    __slots__ = ('data', 'media_type')
    
    def __init__(self, data, media_type):
        """
        Creates a new custom icon type detail.
        
        Parameters
        ----------
        data : `None`, `bytes-like`
            The icon's raw data.
        media_type : `str`
            The icon's media type.
        """
        self.data = data
        self.media_type = media_type
    
    
    @copy_docs(IconDetailsBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' data[length]=')
        repr_parts.append(repr(len(self.data)))
        
        repr_parts.append(', media_type=')
        repr_parts.append(repr(self.media_type))
        
        repr_parts.append('>')
        return ''.join(repr_parts)


class IconType(PreinstancedBase):
    """
    Represents a Discord icon's type.
    
    Attributes
    ----------
    name : `str`
        The name of the icon type.
    value : `int`
        The identifier value the icon type.
    details : ``IconDetailsBase``
        Additional details describing the icon type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``IconType``) items
        Stores the predefined ``IconType``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The icon types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the icon types.
    
    Every predefined icon type can be accessed as class attribute as well:
    
    +-----------------------+---------------+-------+-----------+-------------------+---------------------------------------+
    | Class attribute name  | Name          | Value | Prefix    | Default postfix   | Allowed Postfixes                     |
    +=======================+===============+=======+===========+===================+=======================================+
    | none                  | none          | 0     | `''`      | `''`              | `None`                                |
    +-----------------------+---------------+-------+-----------+-------------------|---------------------------------------+
    | static                | static        | 1     | `''`      | `'png'`           | `'jpg', 'jpeg', 'png', 'webp'`        |
    +-----------------------+---------------+-------+-----------+-------------------|---------------------------------------+
    | animated              | animated      | 2     | `'a_'`    | `'gif'`           | `'jpg', 'jpeg', 'png', 'webp', 'gif'` |
    +-----------------------+---------------+-------+-----------+-------------------+---------------------------------------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ('details',)
    
    
    @classmethod
    @copy_docs(PreinstancedBase._from_value)
    def _from_value(cls, value):
        raise NotImplementedError
    
    
    def __init__(self, value, name, prefix, default_postfix, allowed_postfixes):
        """
        Creates a new icon type with the given parameters and stores it at the type's `.INSTANCES`.
        
        Parameters
        ----------
        value : `str`
            The unique identifier of the icon type.
        name : `str`
            The icon type's name
        prefix : `str`
            Prefix used when building an url with the icon.
        default_postfix : `str`
            Default postfix used when building an url with the icon.
        allowed_postfixes : `None`, `frozenset` of `str`
            The allowed postfixes.
        """
        details = IconDetailsPreinstanced(allowed_postfixes, default_postfix, prefix)
        
        self.details = details
        self.name = name
        self.value = value
        
        self.INSTANCES[value] = self
    
    
    @classmethod
    def from_data(cls, data, name = 'icon'):
        """
        Creates a custom icon type holding actual raw data.
        
        Parameters
        ----------
        data : `bytes-like`
            The icon's raw data.
        name : `str` = `icon`, Optional
            The icon's name.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If `data` is not `bytes-like`.
        """
        if data is None:
            media_type = ''
        
        else:
            if not isinstance(data, (bytes, bytearray, memoryview)):
                raise TypeError(
                    f'`{name}` can be `None`, `bytes-like`, got {data.__class__.__name__}; {reprlib.repr(data)}'
                )
            
            media_type = get_image_media_type(data)
        
        details = IconDetailsCustom(data, media_type)
        
        
        self = PreinstancedBase.__new__(cls)
        self.details = details
        self.name = name
        self.value = -2
        return self
    
    
    def __bool__(self):
        """Returns whether the icon's type is set."""
        if (self.value > 0) or (self.data is not None):
            boolean = True
        else:
            boolean = False
        
        return boolean
    
    
    def __repr__(self):
        """Returns the icon detail's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' value=')
        repr_parts.append(repr(self.value))
        
        repr_parts.append(', name=')
        repr_parts.append(repr(self.name))
        
        repr_parts.append(', details=')
        repr_parts.append(repr(self.details))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @property
    @copy_docs(IconDetailsBase.allowed_postfixes)
    def allowed_postfixes(self):
        return self.details.allowed_postfixes
    
    
    @property
    @copy_docs(IconDetailsBase.default_postfix)
    def default_postfix(self):
        return self.details.default_postfix
    
    
    @property
    @copy_docs(IconDetailsBase.prefix)
    def prefix(self):
        return self.details.prefix
    
    
    @property
    @copy_docs(IconDetailsBase.data)
    def data(self):
        return self.details.data
    
    
    @property
    def base_64_data(self):
        """
        Returns the icon's data in base 64 encoding.
        
        Returns
        -------
        data : `None`, `str`
        """
        data = self.data
        if data is not None:
            return ''.join(['data:', self.media_type, ';base64,', b64encode(data).decode('ascii')])
    
    
    @property
    @copy_docs(IconDetailsBase.media_type)
    def media_type(self):
        return self.details.media_type
    
    
    def allows_postfix(self, postfix):
        """
        Returns whether the icon type allows the given postfix.
        
        Parameters
        ----------
        postfix : `str`
            The postfix to check.
        
        Returns
        -------
        allows_postfix : `bool`
        """
        allowed_postfixes = self.allowed_postfixes
        if (allowed_postfixes is None):
            return False
        
        if (postfix not in allowed_postfixes):
            return False
        
        return True
    
    
    def can_create_url(self):
        """
        Returns whether it is possible to create url with the icon.
        
        Returns
        -------
        can_create_url : `bool`
        """
        return (self.allowed_postfixes is not None)
    
    
    none = P(0, 'none', '', '', None)
    static = P(1, 'static', '', 'png', frozenset(('jpg', 'jpeg', 'png', 'webp')))
    animated = P(2, 'animated', 'a_', 'gif', frozenset(('jpg', 'jpeg', 'png', 'webp', 'gif')))


ICON_TYPE_NONE = IconType.none
ICON_TYPE_STATIC = IconType.static
ICON_TYPE_ANIMATED = IconType.animated


class Icon(RichAttributeErrorBaseType):
    """
    Represents a Discord Icon.
    
    Attributes
    ----------
    hash : `int`
        The icon's hash value.
    type : ``IconType``
        The icon's type.
    """
    __slots__ = ('type', 'hash',)
    
    def __init__(self, icon_type, icon_hash):
        """
        Creates a new ``Icon`` object with the given attributes.
        
        Parameters
        ----------
        icon_type : ``IconType``
            The icon's type.
        icon_hash : `int`
            The icon's hash value.
        """
        self.type = icon_type
        self.hash = icon_hash
    
    
    @property
    def as_base16_hash(self):
        """
        Deprecated and will be removed in 2023 February. Please use ``.as_base_16_hash`` instead.
        """
        warnings.warn(
            (
                f'`{self.__name__}.as_base16_hash` is deprecated and will be removed in 2023 February.'
                f'Please use `.as_base_16_hash` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.as_base_16_hash
    
    
    @property
    def as_base_16_hash(self):
        """
        Returns the discord side representation of the icon.
        
        Returns
        -------
        icon : `None`, `str`
        """
        icon_type = self.type
        if icon_type is ICON_TYPE_NONE:
            icon = None
        
        else:
            icon = self.hash.__format__('0>32x')
            prefix = icon_type.prefix
            if prefix:
                icon = prefix + icon
        
        return icon
    
    
    @property
    def as_base_64_data(self):
        """
        Returns the base 64 version of the icon if applicable.
        
        Returns
        -------
        icon : `None`, `str`
        """
        return self.type.base_64_data
    
    
    hash_info_width = sys.hash_info.width
    if hash_info_width == 32:
        def __hash__(self):
            """Returns the icon's hash."""
            icon_type = self.type
            if icon_type is ICON_TYPE_NONE:
                hash_value = 0
            else:
                icon_hash = self.hash
                hash_value = (
                    (icon_hash >> 96) ^
                    ((icon_hash >> 64) & ((1 << 32) - 1)) ^
                    ((icon_hash >> 32) & ((1 << 32) - 1)) ^
                    (icon_hash & ((1 << 32) - 1))
                )
                
                if icon_type is ICON_TYPE_ANIMATED:
                    hash_value ^= ((1 << 32) - 1)
            
            return hash_value
    
    
    elif hash_info_width == 64:
        def __hash__(self):
            """Returns the icon's hash."""
            icon_type = self.type
            if icon_type is ICON_TYPE_NONE:
                hash_value = 0
            else:
                icon_hash = self.hash
                hash_value = (icon_hash >> 64) ^ (icon_hash & ((1 << 64) - 1))
                if icon_type is ICON_TYPE_ANIMATED:
                    hash_value ^= ((1 << 64) - 1)
            return hash_value
    
    
    else:
        def __hash__(self):
            """Returns the icon's hash."""
            icon_type = self.type
            if icon_type is ICON_TYPE_NONE:
                hash_value = 0
            else:
                hash_value = self.hash
                if icon_type is ICON_TYPE_ANIMATED:
                    hash_value ^= ((1 << 128) - 1)
            
            return hash_value
    
    del hash_info_width
    
    
    def __eq__(self, other):
        """Returns whether the two icons are equal."""
        if other is None:
            if self.type is not ICON_TYPE_NONE:
                return False
            
            if self.hash:
                return False
            
            return True
        
        elif type(self) is type(other):
            # Use `==` to compare value at the case of custom icons.
            if self.type != other.type:
                return False
            
            if self.hash != other.hash:
                return False
            
            return True
        
        elif isinstance(other, tuple):
            if len(other) != 2:
                return NotImplemented
            
            other_type, other_hash = other
            if not isinstance(other_type, IconType):
                return NotImplemented
            
            if not isinstance(other_hash, int):
                return NotImplemented
            
            if self.type != other_type:
                return False
            
            if self.hash != other_hash:
                return False
            
            return True
        
        return NotImplemented
    
    
    def __repr__(self):
        """Returns the representation of the icon."""
        return f'{self.__class__.__name__}(type=ICON_TYPE_{self.type.name.upper()}, hash={self.hash})'
    
    
    def __iter__(self):
        """
        Unpacks the icon.
        
        This method is a generator.
        """
        yield self.type
        yield self.hash
    
    
    def __len__(self):
        """Length hinter (for unpacking if needed)."""
        return 2
    
    
    def __bool__(self):
        return (self.type is not ICON_TYPE_NONE)
    
    
    @classmethod
    def from_base16_hash(cls, icon):
        """
        Converts a discord icon hash value to an ``Icon`` object.
        
        Parameters
        ----------
        icon : `None`, `str`
            Discord icon hash.
        
        Returns
        -------
        self : ``Icon``
        """
        if icon is None:
            icon_type = ICON_TYPE_NONE
            icon_hash = 0
        else:
            if icon.startswith('a_'):
                icon = icon[2:]
                icon_type = ICON_TYPE_ANIMATED
            else:
                icon_type = ICON_TYPE_STATIC
            icon_hash = int(icon, 16)
        
        self = object.__new__(cls)
        self.type = icon_type
        self.hash = icon_hash
        return self


class IconSlot:
    if DOCS_ENABLED:
        __class_doc__ = (
    """
    Internal icon slotter to represent an icon of a discord entity.
    
    Attributes
    ----------
    internal_name : `str`
        The internal name of the icon.
    discord_side_name : `str`
        The discord side name of the icon.
    added_instance_attributes : `tuple` of `str`
        The added instance attribute's name by the icon slot.
    added_class_attributes : `list` of `tuple` (`str`, `Any`)
        The added class attributes by the icon slot.
    
    Class Attributes
    ----------------
    _compile_globals : `dict` of (`str`, `Any`)
        Compile time globals for the generated functions.
    """)
        
        __instance_doc__ = (
    """
    Returns the respective icon.
    
    Returns
    -------
    icon : ``Icon``
    """)
        
        __doc__ = docs_property()
    
    __slots__ = ('internal_name', 'discord_side_name', 'added_instance_attributes', 'added_class_attributes')
    
    _compile_globals = {
        'ICON_TYPE_NONE': ICON_TYPE_NONE,
        'ICON_TYPE_STATIC': ICON_TYPE_STATIC,
        'ICON_TYPE_ANIMATED': ICON_TYPE_ANIMATED,
        'Icon': Icon,
    }
    
    def __new__(cls, internal_name, discord_side_name, url_property, url_as_method, add_updater=True):
        """
        Creates an ``IconSlot`` with the given parameters.
        
        Parameters
        ----------
        internal_name : `str`
            The internal name of the icon.
        discord_side_name : `str`
            The discord side name of the icon.
        url_property : `None`, `function`
            A function what will be used as a property when accessing the icon' url.
        url_as_method : `None`, `function`
            A function what will be used a method when creating a formatted icon url.
        add_updater : `bool` = `True`, Optional
            Whether the icon slot should add updater methods to the class.
        
        Returns
        -------
        self : ``IconSlot``
        """
        added_instance_attribute_name_hash = internal_name + '_hash'
        added_internal_attribute_name_type = internal_name + '_type'
        
        added_class_attributes = []
        if (url_property is not None):
            added_class_attributes.append((f'{internal_name}_url', property(url_property)))
        
        if (url_as_method is not None):
            added_class_attributes.append((f'{internal_name}_url_as', url_as_method))
        
        locals_ = {}
        func_name = f'_set_{internal_name}'
        exec(compile((
            f'def {func_name}(self, data):\n'
            f'    icon = data.get({discord_side_name!r}, None)\n'
            f''
            f'    if icon is None:\n'
            f'        icon_type = ICON_TYPE_NONE\n'
            f'        icon_hash = 0\n'
            f'    else:\n'
            f'        if icon.startswith(\'a_\'):\n'
            f'            icon = icon[2:]\n'
            f'            icon_type = ICON_TYPE_ANIMATED\n'
            f'        else:\n'
            f'            icon_type = ICON_TYPE_STATIC\n'
            f'        icon_hash = int(icon, 16)\n'
            f''
            f'    self.{added_internal_attribute_name_type} = icon_type\n'
            f'    self.{added_instance_attribute_name_hash} = icon_hash\n'
        ), f'<{cls.__name__}>', 'exec', optimize=2), cls._compile_globals, locals_)
        
        added_class_attributes.append((func_name, locals_[func_name]),)
        
        if add_updater:
            locals_ = {}
            func_name = f'_update_{internal_name}'
            exec(compile((
                f'def {func_name}(self, data, old_attributes):\n'
                f'    icon = data.get({discord_side_name!r}, None)\n'
                f''
                f'    if icon is None:\n'
                f'        icon_type = ICON_TYPE_NONE\n'
                f'        icon_hash = 0\n'
                f'    else:\n'
                f'        if icon.startswith(\'a_\'):\n'
                f'            icon = icon[2:]\n'
                f'            icon_type = ICON_TYPE_ANIMATED\n'
                f'        else:\n'
                f'            icon_type = ICON_TYPE_STATIC\n'
                f'        icon_hash = int(icon, 16)\n'
                f''
                f'    self_icon_type = self.{added_internal_attribute_name_type}\n'
                f'    self_icon_hash = self.{added_instance_attribute_name_hash}\n'
                f'    if (self_icon_type is not icon_type) or (self_icon_hash != icon_hash):\n'
                f'        old_attributes[{internal_name!r}] = Icon(self_icon_type, self_icon_hash)\n'
                f'        self.{added_internal_attribute_name_type} = icon_type\n'
                f'        self.{added_instance_attribute_name_hash} = icon_hash\n'
            ), f'<{cls.__name__}>', 'exec', optimize=2), cls._compile_globals, locals_)
            
            added_class_attributes.append((func_name, locals_[func_name]),)
        
        self = object.__new__(cls)
        self.internal_name = internal_name
        self.discord_side_name = discord_side_name
        self.added_instance_attributes = (added_internal_attribute_name_type, added_instance_attribute_name_hash)
        self.added_class_attributes = added_class_attributes
        return self
    
    
    def __set_slot__(self, attribute_name, class_attributes, class_slots):
        """Applies the changes of the icon slot on the class's attributes."""
        
        # Extend the slots of the class
        class_slots.update(self.added_instance_attributes)
        
        # Add the extra class attributes to the class
        for name, value in self.added_class_attributes:
            class_attributes[name] = value
    
    
    def __get__(self, attribute, type_):
        """Returns self if called from class, meanwhile an ``Icon`` if called from an object."""
        if attribute is None:
            return self
        
        icon_type_name, icon_hash_name = self.added_instance_attributes
        icon_type = getattr(attribute, icon_type_name)
        icon_hash = getattr(attribute, icon_hash_name)
        return Icon(icon_type, icon_hash)
    
    
    def __set__(self, instance, icon):
        """Can't set attribute."""
        if (icon is None):
            icon_type = ICON_TYPE_NONE
            icon_hash = 0
        
        elif isinstance(icon, Icon):
            icon_type, icon_hash = icon
        
        else:
            raise TypeError(
                f'`{instance.__class__.__name__}.{self.internal_name}` can only be `None`, `{Icon.__name__}`, '
                f'got {icon.__class__.__name__}; {icon!r}.'
            )
        
        icon_type_name, icon_hash_name = self.added_instance_attributes
        setattr(instance, icon_type_name, icon_type)
        setattr(instance, icon_hash_name, icon_hash)
    
    
    def __delete__(self, instance):
        """Can't delete attribute."""
        raise AttributeError(
            f'Can\'t delete `{instance.__class__.__name__}.{self.internal_name}`.'
        )
    
    
    def validate_data_icon(self, icon):
        """
        Validates the given icon data.
        
        Parameters
        ----------
        icon : `None`, `bytes`, `bytearray`, `memoryview`, ``Icon``
            The icon to validate.
        
        Returns
        -------
        icon : `None`, ``Icon``
            The validated icon.
        
        Raises
        ------
        TypeError
            If `icon` field's type is incorrect.
        ValueError
            if `icon` has unexpected media type.
        """
        if icon is None:
            return None
        
        if isinstance(icon, Icon):
            return icon
        
        icon_type = IconType.from_data(icon)
        if icon_type and (not icon_type.media_type):
            raise ValueError(
                f'`{self.internal_name}` received unknown media type; Got {reprlib.repr(icon_type.data)}.'
            )
        
        icon_hash = 0
        
        return Icon(icon_type, icon_hash)
    
    
    def parse_data_from_keyword_parameters(self, keyword_parameters):
        """
        Parses data from keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `Any`) items
            Keyword parameters passed to parse the icon from.
        
        Returns
        -------
        icon : `None`, ``Icon``
            The parsed out icon.
        
        Raises
        ------
        TypeError
            If `icon` field's type is incorrect.
        ValueError
            if `icon` has unexpected media type.
        """
        try:
            icon = keyword_parameters.pop(self.internal_name)
        except KeyError:
            return None
        
        return self.validate_data_icon(icon)
    
    
    def parse_from_keyword_parameters(self, keyword_parameters, *, allow_data = False):
        """
        Parses the icon out from the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `Any`) items
            Keyword parameters passed to parse the icon from.
        allow_data : `bool` = `False`, Optional (Keyword only)
            Whether data parsing is allowed.
        
        Returns
        -------
        icon : `None`, ``Icon``
            The parsed out icon.
        
        Raises
        ------
        TypeError
            If any of expected value's type is invalid.
        ValueError
            If any of the expected value's type is valid, but it's value is not.
        """
        icon_type_name, icon_hash_name = self.added_instance_attributes
        try:
            icon = keyword_parameters.pop(self.internal_name)
        except KeyError:
            try:
                icon_hash = keyword_parameters.pop(icon_hash_name)
            except KeyError:
                return None
            
            if type(icon_hash) is int:
                pass
            
            elif isinstance(icon_hash, int):
                icon_hash = int(icon_hash)
            
            else:
                raise TypeError(
                    f'`{icon_hash_name}` can be `int`, got {icon_hash.__class__.__name__}; {icon_hash!r}.'
                )
            
            if icon_hash < 0 or icon_hash > ((1 << 128) - 1):
                raise ValueError(
                    f'`{icon_hash_name}` cannot be negative or longer than 128 bits, got {icon_hash!r}.'
                )
            
            try:
                icon_type = keyword_parameters.pop(icon_type_name)
            except KeyError:
                if icon_hash == 0:
                    icon_type = ICON_TYPE_NONE
                else:
                    icon_type = ICON_TYPE_STATIC
            else:
                if (type(icon_type) is not IconType):
                    raise TypeError(
                        f'`{icon_type_name}` can be `{IconType.__name__}`, got '
                        f'{icon_type.__class__.__name__}; {icon_type!r}.'
                    )
                
                if (icon_type is ICON_TYPE_NONE) and icon_hash:
                    raise ValueError(
                        f'If `{icon_type_name}` is passed as `ICON_TYPE_NONE`, then `{icon_hash_name}` '
                        f'can be passed only as `0`, meanwhile got `{icon_hash}`.'
                    )
            
        else:
            if icon is None:
                icon_type = ICON_TYPE_NONE
                icon_hash = 0
            
            elif type(icon) is Icon:
                icon_type = icon.type
                icon_hash = icon.hash
            
            elif isinstance(icon, str):
                if icon.startswith('a_'):
                    icon = icon[2:]
                    icon_type = ICON_TYPE_ANIMATED
                else:
                    icon_type = ICON_TYPE_STATIC
                icon_hash = int(icon, 16)
            
            elif allow_data and isinstance(icon, (bytes, bytearray, memoryview)):
                icon_type = IconType.from_data(icon)
                icon_hash = 0
            
            else:
                raise TypeError(
                    f'`{self.internal_name}` can be passed as `None`, `{Icon.__name__}`, `str` '
                    f'(or bytes-like if allowed), got {icon.__class__.__name__}; {reprlib.repr(icon)}.'
                )
        
        return Icon(icon_type, icon_hash)
    
    
    def preconvert(self, keyword_parameters, processable):
        """
        Used at preconverters to parse out from the passed keyword_parameters the icon of the entity.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `Any`) items
            Keyword parameters passed to the respective preconverter.
        processable : `list` of `tuple` (`str`, `Any`)
            A list of instance attributes which will be set when all the passed keyword_parameters are validated.
        
        Raises
        ------
        TypeError
            If any of expected value's type is invalid.
        ValueError
            If any of the expected value's type is valid, but it's value is not.
        """
        icon = self.parse_from_keyword_parameters(keyword_parameters)
        if (icon is None):
            return
        
        icon_type, icon_hash = icon
        icon_type_name, icon_hash_name = self.added_instance_attributes
        
        processable.append((icon_type_name, icon_type))
        processable.append((icon_hash_name, icon_hash))
    
    
    def put_into(self, icon, data, defaults, *, as_data = False):
        """
        Puts the icon into the given data.
        
        Parameters
        ----------
        icon : ``Icon``
            The icon to serialise.
        data : `dict` of (`str`, `Any`) items
            The data put the icon into.
        defaults : `bool`
            Whether the `icon`'s value should be put even if it is the default value.
        as_data : `bool` = `False`, Optional (Keyword only)
            Whether we want the icon is data or in hash form.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        if icon is None:
            field_value = None
        
        else:
            if as_data:
                field_value = icon.as_base_64_data
            else:
                field_value = icon.as_base_16_hash
        
        if defaults or (field_value is not None):
            data[self.discord_side_name] = field_value
        
        return data



__all__ = ()

from ...field_parsers import bool_parser_factory, force_string_parser_factory, nullable_string_parser_factory
from ...field_putters import bool_optional_putter_factory, force_string_putter_factory, nullable_string_putter_factory
from ...field_validators import (
    bool_validator_factory, force_string_validator_factory, nullable_string_validator_factory
)

from .constants import DESCRIPTION_LENGTH_MAX, LABEL_LENGTH_MAX, VALUE_LENGTH_MAX


# default

parse_default = bool_parser_factory('default', False)
put_default_into = bool_optional_putter_factory('default', False)
validate_default = bool_validator_factory('default')

# description

parse_description = nullable_string_parser_factory('description')
put_description_into = nullable_string_putter_factory('description')
validate_description = nullable_string_validator_factory('description', 0, DESCRIPTION_LENGTH_MAX)

# label

parse_label = force_string_parser_factory('label')
put_label_into = force_string_putter_factory('label')
validate_label = nullable_string_validator_factory('label', 0, LABEL_LENGTH_MAX)

# value

parse_value = force_string_parser_factory('value')
put_value_into = force_string_putter_factory('value')
validate_value = force_string_validator_factory('value', 0, VALUE_LENGTH_MAX)



__all__ = ('ComponentMetadataButton', )

import reprlib, warnings

from scarletio import copy_docs

from ...utils import url_cutter

from ..shared_fields import (
    parse_custom_id, parse_emoji, put_custom_id_into, put_emoji_into, validate_custom_id, validate_emoji
)
from ..shared_helpers import create_auto_custom_id

from .base import ComponentMetadataBase
from .constants import BUTTON_STYLE_DEFAULT
from .fields import (
    parse_button_style, parse_enabled, parse_label, parse_url, put_button_style_into, put_enabled_into, put_label_into,
    put_url_into, validate_button_style, validate_enabled, validate_label, validate_url
)
from .preinstanced import ButtonStyle


class ComponentMetadataButton(ComponentMetadataBase):
    """
    Button component metadata.
    
    Attributes
    ----------
    button_style : ``ButtonStyle``
        The button's style.
    
    custom_id : `None`, `str`
        Custom identifier to detect which button was clicked by the user.
        
        > Mutually exclusive with the `url` field.
    
    emoji : `None`, ``Emoji``
        Emoji of the button if applicable.
    
    enabled : `bool`
        Whether the component is enabled.
    
    label : `None`, `str`
        Label of the component.
    
    url : `None`, `str`
        Url to redirect to when clicking on the button.
        
        > Mutually exclusive with the `custom_id` field.
    """
    __slots__ = ('button_style', 'custom_id', 'emoji', 'enabled', 'label', 'url',)
    
    @copy_docs(ComponentMetadataBase.__new__)
    def __new__(cls, keyword_parameters):
        # button_style
        try:
            button_style = keyword_parameters.pop('button_style')
        except KeyError:
            button_style = BUTTON_STYLE_DEFAULT
        else:
            button_style = validate_button_style(button_style)
        
        # custom_id
        try:
            custom_id = keyword_parameters.pop('custom_id')
        except KeyError:
            custom_id = None
        else:
            custom_id = validate_custom_id(custom_id)
        
        # emoji
        try:
            emoji = keyword_parameters.pop('emoji')
        except KeyError:
            emoji = None
        else:
            emoji = validate_emoji(emoji)
        
        # enabled
        try:
            enabled = keyword_parameters.pop('enabled')
        except KeyError:
            enabled = True
        else:
            enabled = validate_enabled(enabled)
        
        # label
        try:
            label = keyword_parameters.pop('label')
        except KeyError:
            label = None
        else:
            label = validate_label(label)
        
        # url
        try:
            url = keyword_parameters.pop('url')
        except KeyError:
            url = None
        else:
            url = validate_url(url)
        
        
        self = object.__new__(cls)
        self.button_style = button_style
        self.custom_id = custom_id
        self.emoji = emoji
        self.enabled = enabled
        self.label = label
        self.url = url
        
        self._validate()
        
        return self
    
    
    def _validate(self):
        """
        Runs additional validation after the component is created.
        
        Raises
        ------
        ValueError
            - `url` is mutually exclusive with `custom_id`.
        """
        if self.url is None:
            if self.custom_id is None:
                self.custom_id = create_auto_custom_id()
            
            if (self.button_style is ButtonStyle.link) or (self.button_style is ButtonStyle.none):
                self.button_style = BUTTON_STYLE_DEFAULT
            
        else:
            if self.custom_id is None:
                self.button_style = ButtonStyle.link
            
            else:
                raise ValueError(
                    f'`custom_id` and `url` fields are mutually exclusive, got '
                    f'custom_id = {self.custom_id!r}, url = {self.url!r}.'
                )
    
    
    @copy_docs(ComponentMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # Descriptive fields : button_style
        
        # button_style
        button_style = self.button_style
        repr_parts.append(' button_style = ')
        repr_parts.append(button_style.name)
        repr_parts.append(' (')
        repr_parts.append(repr(button_style.value))
        repr_parts.append(')')
        
        # System fields : custom_id
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            repr_parts.append(', custom_id = ')
            repr_parts.append(reprlib.repr(custom_id))
        
        # Text fields : emoji & label
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            repr_parts.append(', emoji = ')
            repr_parts.append(repr(emoji))
        
        # label
        label = self.label
        if (label is not None):
            repr_parts.append(', label = ')
            repr_parts.append(reprlib.repr(label))
        
        
        # Optional descriptive fields: url & enabled
        
        # url
        url = self.url
        if (url is not None):
            repr_parts.append(', url = ')
            repr_parts.append(url_cutter(url))
        
        # enabled
        enabled = self.enabled
        if (not enabled):
            repr_parts.append(', enabled = ')
            repr_parts.append(repr(enabled))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # button_style
        hash_value ^= self.button_style.value
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # emoji
        emoji = self.emoji
        if (emoji is not None):
            hash_value ^= emoji.id
        
        # enabled
        if self.enabled:
            hash_value ^= 1 << 8
        
        # label
        label = self.label
        if (label is not None):
            hash_value ^= hash(label)
        
        # url
        url = self.url
        if (url is not None):
            hash_value ^= hash(url)
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # button_style
        if self.button_style is not other.button_style:
            return False
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # emoji
        if self.emoji is not other.emoji:
            return False
        
        # enabled
        if self.enabled != other.enabled:
            return False
        
        # label
        if self.label != other.label:
            return False
            
        # url
        if self.url != other.url:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        
        self.button_style = parse_button_style(data)
        self.custom_id = parse_custom_id(data)
        self.emoji = parse_emoji(data)
        self.enabled = parse_enabled(data)
        self.label = parse_label(data)
        self.url = parse_url(data)
        
        return self
    
    
    @copy_docs(ComponentMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data =  {}
        
        put_button_style_into(self.button_style, data, defaults)
        put_custom_id_into(self.custom_id, data, defaults)
        put_emoji_into(self.emoji, data, defaults)
        put_enabled_into(self.enabled, data, defaults)
        put_label_into(self.label, data, defaults)
        put_url_into(self.url, data, defaults)
        
        return data
    
    
    @copy_docs(ComponentMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        new.button_style = self.button_style
        new.custom_id = self.custom_id
        new.emoji = self.emoji
        new.enabled = self.enabled
        new.label = self.label
        new.url = self.url
        
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy_with)
    def copy_with(self, keyword_parameters):
        # button_style
        try:
            button_style = keyword_parameters.pop('button_style')
        except KeyError:
            button_style = self.button_style
        else:
            button_style = validate_button_style(button_style)
        
        # custom_id
        try:
            custom_id = keyword_parameters.pop('custom_id')
        except KeyError:
            custom_id = self.custom_id
        else:
            custom_id = validate_custom_id(custom_id)
        
        # emoji
        try:
            emoji = keyword_parameters.pop('emoji')
        except KeyError:
            emoji = self.emoji
        else:
            emoji = validate_emoji(emoji)
        
        # enabled
        try:
            enabled = keyword_parameters.pop('enabled')
        except KeyError:
            enabled = self.enabled
        else:
            enabled = validate_enabled(enabled)
        
        # label
        try:
            label = keyword_parameters.pop('label')
        except KeyError:
            label = self.label
        else:
            label = validate_label(label)
        
        # url
        try:
            url = keyword_parameters.pop('url')
        except KeyError:
            url = self.url
        else:
            url = validate_url(url)
        
        # Deprecated: style
        
        try:
            style = keyword_parameters.pop('style')
        except KeyError:
            pass
        else:
            warnings.warn(
                (
                    '`style` parameter of components is deprecated and will be removed in 2023 February. '
                    'Please use `button_style` for button components.'
                ),
                FutureWarning,
                stacklevel = 3,
            )
            
            button_style = validate_button_style(style)
        
        # Construct
        
        self = object.__new__(type(self))
        self.button_style = button_style
        self.custom_id = custom_id
        self.emoji = emoji
        self.enabled = enabled
        self.label = label
        self.url = url
        
        self._validate()
        
        return self



__all__ = ()

from .color import Color


def preconvert_snowflake(snowflake, name):
    """
    Converts the given `snowflake` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    snowflake : `str`, `int`
        The snowflake to convert.
    name : `str`
        The name of the snowflake.
    
    Returns
    -------
    snowflake : `int`
    
    Raises
    ------
    TypeError
        - If `snowflake` was not passed neither as `int`, `str`.
    ValueError
        - If `snowflake` was passed as `str` and cannot be converted to `int`.
        - If the converted `snowflake` is negative or it's bit length is over 64.
    """
    snowflake_type = type(snowflake)
    if snowflake_type is int:
        pass
    if issubclass(snowflake_type, int):
        snowflake = int(snowflake)
    # JSON uint64 is str
    elif issubclass(snowflake_type, str):
        if 6 < len(snowflake) < 21 and snowflake.isdigit():
            snowflake = int(snowflake)
        else:
            raise ValueError(
                f'`{name}` can be `int`, `str`, got `str`, but not a valid'
                f'snowflake (7-20 length, digit only), got {snowflake!r}.'
            )
    else:
        raise TypeError(
            f'`{name}` can be `int`, `str`, got {snowflake_type.__name__}; {snowflake!r}.'
        )
    
    if snowflake < 0 or snowflake > ((1 << 64) - 1):
        raise ValueError(
            f'`{name}` can be only uint64, got {snowflake!r}.'
        )
    
    return snowflake


def preconvert_snowflake_array(snowflake_array, name):
    """
    Converts the given `snowflake_array` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    snowflake_array : `None` or `iterable` of (`str`, `int`)
        The snowflakes to convert.
    name : `str`
        The name of the snowflake array.
    
    Returns
    -------
    snowflake_array : `None`, `tuple` of `int`
        The returned value is always sorted.
    
    Raises
    ------
    TypeError
        - If `snowflake_array` is neither `None`, `iterable`
        - If `snowflake_array` contains a non `int` nor `str`.
    ValueError
        - If `snowflake_array`contains a `str`, what cannot be converted to `int`.
        - If a converted `snowflake` is negative or it's bit length is over 64.
    """
    if (snowflake_array is None):
        return None
    
    snowflake_array_processed = None
    
    if getattr(preconvert_snowflake_array, '__iter__', None) is None:
        raise TypeError(
            f'`{name}` can be `list`, `tuple`, `set`, got '
            f'{snowflake_array.__class__.__name__}; {snowflake_array!r}.'
        )
    
    for snowflake in snowflake_array:
        snowflake_type = type(snowflake)
        if snowflake_type is int:
            pass
        if issubclass(snowflake_type, int):
            snowflake = int(snowflake)
        # JSON uint64 is str
        elif issubclass(snowflake_type, str):
            if 6 < len(snowflake) < 21 and snowflake.isdigit():
                snowflake = int(snowflake)
            else:
                raise ValueError(
                    f'`{name}`\'s can contain `int`, `str` elements, got `str`, but not a valid snowflake '
                    f'(7-20 length, digit only), got {snowflake!r}; snowflake_array={snowflake_array!r}.'
                )
        else:
            raise TypeError(
                f'`{name}`\'s can contain `int`, `str` elements, got '
                f'{snowflake_type.__name__}; {snowflake!r}; snowflake_array={snowflake_array!r}.'
            )
        
        if snowflake < 0 or snowflake > ((1 << 64) - 1):
            raise ValueError(
                f'`{name}`\'s elements can be only uint64, got '
                f'{snowflake!r}; snowflake_array={snowflake_array!r}.'
            )
        
        if snowflake_array_processed is None:
            snowflake_array_processed = set()
        
        snowflake_array_processed.add(snowflake)
    
    if (snowflake_array_processed is not None):
        snowflake_array = tuple(sorted(snowflake_array_processed))

    return snowflake_array


def preconvert_discriminator(discriminator):
    """
    Converts the given `discriminator` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    discriminator : `str`, `int`
        The discriminator of an user to convert.
    
    Returns
    -------
    discriminator : `int`
    
    Raises
    ------
    TypeError
        If `discriminator` was not passed neither as `int`, `str`.
    ValueError
        - If `discriminator` was passed as `str` and it is not numerical or it's length is over `4`.
        - If the `discriminator`'s value is less than `0` or is over than `9999`.
    """
    if (type(discriminator) is int):
        pass
    elif isinstance(discriminator, int):
        discriminator = int(discriminator)
    # Discord sends `discriminator` as `str`, so lets accept that as well.
    elif isinstance(discriminator, str):
        if 0 < len(discriminator) < 5 and discriminator.isdigit():
            raise ValueError(
                f'`discriminator` was given as a `str`, but it is either not numeric, or it\'s length is not 1-4, got '
                f'{discriminator!r}.'
            )
        discriminator = int(discriminator)
    
    else:
        raise TypeError(
            f'`discriminator` can be `int`, `str`, got '
            f'{discriminator.__class__.__name__}; {discriminator!r}.'
        )
    
    if discriminator < 0 or discriminator > 9999:
        raise ValueError(
            f'`discriminator` can be between 0 and 9999, got got {discriminator!r}.'
        )
    
    return discriminator


def preconvert_color(color, name, nullable):
    """
    Converts the given `color` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    color : `int`
        The color to convert.
    name : `str`
        The name of the value.
    nullable : `bool`
        Whether the value is nullable.
    
    Returns
    -------
    color : `Color`
    
    Raises
    ------
    TypeError
        If `color` was not passed as `int`.
    ValueError
        If the `color`'s value is less than `0` or is over than `0xffffff`.
    """
    if not (nullable and (color is None)):
        if type(color) is Color:
            pass
        elif isinstance(color, int):
            color = Color(color)
        else:
            raise TypeError(
                f'`{name}` can be {"`None`, " if nullable else ""}`{Color.__name__}`, `int`, '
                f'got {color.__class__.__name__}; {color!r}.'
            )
            
            if color < 0 or color > 0xffffff:
                raise ValueError(
                    f'`{name}` can be between 0 and 0xffffff, got {color!r}.'
                )
    
    return color


def preconvert_str(value, name, lower_limit, upper_limit):
    """
    Converts the given `value` to an acceptable string by the wrapper.
    
    Parameters
    ----------
    value : `str`
        The string to convert,
    name : `str`
        The name of the value.
    lower_limit : `int`
        The minimal length of the string.
    upper_limit : `int`
        The maximal length of the string.
    
    Returns
    -------
    value : `str`
    
    Raises
    ------
    TypeError
        If `value` was not passed as `str`.
    ValueError
        If the `value`'s length is less than the given `lower_limit` or is higher than the given than the given
        `upper_limit`.
    """
    if type(value) is str:
        pass
    elif isinstance(value, str):
        value = str(value)
    else:
        raise TypeError(
            f'`{name}` can be `str`, got {value.__class__.__name__}; {value!r}.'
        )
    
    length = len(value)
    if (length != 0) and (length < lower_limit or length > upper_limit):
        raise ValueError(
            f'`{name}` can be between length {lower_limit} and {upper_limit}, got {length!r}; {value!r}.'
        )
    
    return value


def preconvert_iterable_of_str(value, name, iterable_lower_limit, iterable_upper_limit, lower_limit, upper_limit):
    """
    Converts the given `value` to an acceptable iterable of string by the wrapper.
    
    Parameters
    ----------
    value : `str`
        The string to convert,
    name : `str`
        The name of the value.
    iterable_lower_limit : `int`
        The minimal length of `value`.
    iterable_upper_limit : `int`
        The maximal length of `value`.
    lower_limit : `int`
        The minimal length of the string.
    upper_limit : `int`
        The maximal length of the string.
    
    Returns
    -------
    converted_value : `set` of `str`
    
    Raises
    ------
    TypeError
        - If `value` was not passed as an `iterable`.
        - If `value` contains a non `str` element.
    ValueError
        - If the `value`'s length is less than the given `iterable_lower_limit` or is higher than the given than
            the given `iterable_upper_limit`.
        - If `value` contains a string whats length is less than `lower_limit` or is more than `upper_limit`.
    """
    iterator = getattr(type(value), '__iter__', None)
    if iterator is None:
        raise TypeError(
            f'`{name}` can be `iterable`, got {value.__class__.__name__}; {value!r}.'
        )
    
    converted_value = set()
    
    for value_element in iterator(value):
        if type(value_element) is str:
            pass
        elif isinstance(value_element, str):
            value_element = str(value_element)
        else:
            raise TypeError(
                f'`{name}` can contain `str` elements, got {value_element.__class__.__name__}; {value_element!r}; '
                f'value={value!r}.'
            )
        
        length = len(value_element)
        if length == 0:
            continue
        
        if (length < lower_limit) or (length > upper_limit):
            raise ValueError(
                f'`{name}` can contains elements between length {lower_limit} and {upper_limit}, '
                f'got {length!r} {value_element!r}; value={value!r}.'
            )
        
        converted_value.add(value_element)
    
    length = len(converted_value)
    if (length != 0) and (length < iterable_lower_limit or length > iterable_upper_limit):
        raise ValueError(
            f'`{name} can be in between length {iterable_lower_limit} and {iterable_upper_limit}, '
            f'got {length!r}; {value!r}.'
        )
    
    return converted_value


def preconvert_bool(value, name):
    """
    Converts the given `value` to an acceptable boolean by the wrapper.
    
    Parameters
    ----------
    value : `int`
        The value to convert.
    name : `str`
        The name of the value.
    
    Returns
    -------
    value : `str`
    
    Raises
    ------
    TypeError
        If `value` was not passed as `int`.
    ValueError
        If value was passed as an `int`, but not as `0`, `1` either.
    """
    if (type(value) is bool):
        pass
    elif isinstance(value, int):
        if (value not in (0, 1)):
            raise ValueError(
                f'`{name}` was given as `int`, but neither as `0`, `1`, got {value!r}.'
            )
        value = bool(value)
    else:
        raise TypeError(
            f'`{name}` can be `bool`, `int` as `0`, `1`, got {value.__class__.__name__}; {value!r}.'
        )
    
    return value


def preconvert_flag(flag, name, type_):
    """
    Converts the given `flag` to an acceptable flag by the wrapper.
    
    Parameters
    ----------
    flag : `int`
        The flag to convert.
    name : `str`
        The name of the flag.
    type_ : `int` subtype
        The type of the output flag.
    
    Returns
    -------
    flag : `type_`
    
    Raises
    ------
    TypeError
        If `flag` was not given as `int`.
    ValueError
        If `flag` was given as `int`, but it's value is less than `0` or it's bit length is over `64`.
    """
    flag_type = flag.__class__
    if (flag_type is type_):
        pass
    elif issubclass(flag_type, int):
        flag = type_(flag)
    else:
        raise TypeError(
            f'`{name}` can be `{type_.__name__}`, got {flag_type.__name__}; {flag!r}.'
        )
    
    if flag < 0 or flag > ((1 << 64) - 1):
        raise ValueError(
            f'`{name}` can be only uint64, got {flag!r}.'
        )
    
    return flag


def preconvert_preinstanced_type(value, name, type_):
    """
    Converts the given `value` to an acceptable value by the wrapper.
    
    Parameters
    ----------
    value : `Any`
        The value to convert.
    name : `str`
        The name of the value.
    type_ : ``PreinstancedBase``
        The preinstanced type.
    
    Returns
    -------
    value : ``PreinstancedBase``
    
    Raises
    ------
    TypeError
        If `value` was not given as `type_`, neither as `type_.value`'s type's instance.
    ValueError
        If there is no preinstanced object for the given `value`.
    """
    value_type = value.__class__
    if (value_type is not type_):
        value_expected_type = type_.VALUE_TYPE
        if value is None:
            value = type_.get(value)
        
        elif value_type is value_expected_type:
            pass
        
        elif issubclass(value_type, value_expected_type):
            value = value_expected_type(value)
        
        else:
            raise TypeError(
                f'`{name}` can be `{type_.__name__}`, `{value_expected_type.__name__}` , got '
                f'{value_type.__name__}; {value!r}.'
            )
        
        try:
            value = type_.INSTANCES[value]
        except LookupError:
            raise ValueError(
                f'There is no predefined `{name}` for value: {value!r}.'
            ) from None
    
    return value


def _pre_validate_int(value, name):
    """
    Converts the given `value` to `int`.

    Parameters
    ----------
    value : `Any`
        The value to convert.
    name : `str`
        The name of the value.
    
    Returns
    -------
    value : `int`
    
    Raises
    ------
    TypeError
        If `value` was not given as `int`.
    """
    if type(value) is int:
        pass
    elif isinstance(value, int):
        value = int(value)
    else:
        raise TypeError(
            f'`{name}` can be `int`, got {value.__class__.__name__}; {value!r}.'
        )

    return value


def preconvert_int(value, name, lower_limit, upper_limit):
    """
    Converts the given `value` to an acceptable integer by the wrapper.
    
    Parameters
    ----------
    value : `Any`
        The value to convert.
    name : `str`
        The name of the value.
    lower_limit : `int`
        The minimal value of `value`.
    upper_limit : `int`
        The maximal value of `value`.
    
    Returns
    -------
    value : `int`
    
    Raises
    ------
    TypeError
        If `value` was not given as `int`.
    ValueError
        If `value` is less than `lower_limit`, or is higher than the `upper_limit`.
    """
    value = _pre_validate_int(value, name)
    
    if (value < lower_limit) or (value > upper_limit):
        raise ValueError(
            f'`{name}` can be between {lower_limit} and {upper_limit}, got {value!r}.'
        )
    
    return value


def preconvert_int_options(value, name, options):
    """
    Converts the given `value` to an acceptable integer by the wrapper.
    
    Parameters
    ----------
    value : `Any`
        The value to convert.
    name : `str`
        The name of the value.
    options : `frozenset`
        The options, from which `value` should be one.
    
    Returns
    -------
    value : `int`
    
    Raises
    ------
    TypeError
        If `value` was not given as `int`.
    ValueError
        If `value` is less than `lower_limit`, or is higher than the `upper_limit`.
    """
    value = _pre_validate_int(value, name)
    
    if value not in options:
        raise ValueError(
            f'`{name}` can be any of: {", ".join(str(option) for option in options)}, got {value!r}.'
        )
    
    return value


def preconvert_float(value, name, lower_limit, upper_limit):
    """
    Converts the given `value` to an acceptable float by the wrapper.
    
    Parameters
    ----------
    value : `Any`
        The value to convert.
    name : `str`
        The name of the value.
    lower_limit : `float`
        The minimal value of `value`.
    upper_limit : `float`
        The maximal value of `value`.
    
    Returns
    -------
    value : `float`
    
    Raises
    ------
    TypeError
        If `value` was not given as `float`.
    ValueError
        If `value` is less than `lower_limit`, or is higher than the `upper_limit`.
    """
    if type(value) is float:
        pass
    elif isinstance(value, float):
        value = int(value)
    else:
        float_converter = getattr(type(value), '__float__', None)
        if float_converter is None:
            raise TypeError(
                f'`{name}` can be `float`, got {value.__class__.__name__}; {value!r}.'
            )
        
        value = float_converter(value)
    
    if (value < lower_limit) or (value > upper_limit):
        raise ValueError(
            f'`{name}` can be between {lower_limit} and {upper_limit}, got {value!r}.'
        )
    
    return value


def get_type_names(type_or_types):
    """
    Gets the given type(s)'s name closed within \` characters.
    
    Parameters
    ----------
    type_or_types : `type`, `tuple` of `type`
        The type(s) to get name of.
    
    Returns
    -------
    type_names : `str`
    """
    type_name_parts = []
    
    if isinstance(type_or_types, type):
        type_name_parts.append('`')
        type_name_parts.append(type_or_types.__name__)
        type_name_parts.append('`')
    else:
        length = len(type_or_types)
        if length:
            index = 0
            while True:
                type_ = type_or_types[index]
                index += 1
                
                type_name_parts.append('`')
                type_name_parts.append(type_.__name__)
                type_name_parts.append('`')
                
                if index == length:
                    break
                
                type_name_parts.append(', ')
                continue
    
    return ''.join(type_name_parts)



__all__ = ('ActivityAssets',)

from scarletio import copy_docs

from .base import ActivityFieldBase


def _assert__activity_secrets__image_large(image_large):
    """
    Asserts the `image_large` parameter of ``ActivitySecrets.__new__`` method.
    
    Parameters
    ----------
    image_large : `None`, `str`
        The id of the activity's large asset to display.
    
    Raises
    ------
    AssertionError
        - If `image_large` is not `None`, `str`.
    """
    if (image_large is not None) and (not isinstance(image_large, str)):
        raise AssertionError(
            f'`image_large` can be `None`, `str`, got {image_large.__class__.__name__}; {image_large!r}.'
        )
    
    return True


def _assert__activity_secrets__image_small(image_small):
    """
    Asserts the `image_small` parameter of ``ActivitySecrets.__new__`` method.
    
    Parameters
    ----------
    image_small : `None`, `str`
        The id of the activity's small asset to display.
    
    Raises
    ------
    AssertionError
        - If `image_small` is not `None`, `str`.
    """
    if (image_small is not None) and (not isinstance(image_small, str)):
        raise AssertionError(
            f'`image_small` can be `None`, `str`, got {image_small.__class__.__name__}; {image_small!r}.'
        )
    
    return True


def _assert__activity_secrets__text_large(text_large):
    """
    Asserts the `text_large` parameter of ``ActivitySecrets.__new__`` method.
    
    Parameters
    ----------
    text_large : `None`, `str`
        The hover text of the large asset.
    
    Raises
    ------
    AssertionError
        - If `text_large` is not `None`, `str`.
    """
    if (text_large is not None) and (not isinstance(text_large, str)):
        raise AssertionError(
            f'`text_large` can be `None`, `str`, got {text_large.__class__.__name__}; {text_large!r}.'
        )
    
    return True


def _assert__activity_secrets__text_small(text_small):
    """
    Asserts the `text_small` parameter of ``ActivitySecrets.__new__`` method.
    
    Parameters
    ----------
    text_small : `None`, `str`
        The hover text of the small asset.
    
    Raises
    ------
    AssertionError
        - If `text_small` is not `None`, `str`.
    """
    if (text_small is not None) and (not isinstance(text_small, str)):
        raise AssertionError(
            f'`text_small` can be `None`, `str`, got {text_small.__class__.__name__}; {text_small!r}.'
        )
    
    return True


class ActivityAssets(ActivityFieldBase):
    """
    Represents a discord activity asset.
    
    Attributes
    ----------
    image_large : `None`, `str`
        The id of the activity's large asset to display.
    image_small : `None`, `str`
        The id of the activity's small asset to display.
    text_large : `None`, `str`
        The hover text of the large asset.
    text_small : `None`, `str`
        The hover text of the small asset.
    """
    __slots__ = ('image_large', 'image_small', 'text_large', 'text_small',)
    
    def __new__(cls, *, image_large=None, image_small=None, text_large=None, text_small=None):
        """
        Creates a new ``ActivityAssets`` from the given parameters.
        
        Parameters
        ----------
        image_large : `None`, `str` = `None`, Optional (Keyword only)
            The id of the activity's large asset to display.
        image_small : `None`, `str` = `None`, Optional (Keyword only)
            The id of the activity's small asset to display.
        text_large : `None`, `str` = `None`, Optional (Keyword only)
            The hover text of the large asset
        text_small : `None`, `str` = `None`, Optional (Keyword only)
            The hover text of the small asset.
        
        Raises
        ------
        AssertionError
            - If `image_large` is neither `None` nor `str`.
            - If `image_small` is neither `None` nor `str`.
            - If `text_large` is neither `None` nor `str`.
            - If `text_small` is neither `None` nor `str`.
        """
        assert _assert__activity_secrets__image_large(image_large)
        assert _assert__activity_secrets__image_small(image_small)
        assert _assert__activity_secrets__text_large(text_large)
        assert _assert__activity_secrets__text_small(text_small)
        
        if (image_large is not None) and (not image_large):
            image_large = None
            
        if (image_small is not None) and (not image_small):
            image_small = None
            
        if (text_large is not None) and (not text_large):
            text_large = None
        
        if (text_small is not None) and (not text_small):
            text_small = None
        
        self = object.__new__(cls)
        self.image_large = image_large
        self.image_small = image_small
        self.text_large = text_large
        self.text_small = text_small
        return self
    
    
    @classmethod
    @copy_docs(ActivityFieldBase.from_data)
    def from_data(cls, assets_data):
        self = object.__new__(cls)
        self.image_large = assets_data.get('large_image', None)
        self.image_small = assets_data.get('small_image', None)
        self.text_large = assets_data.get('large_text', None)
        self.text_small = assets_data.get('small_text', None)
        return self
    
    
    @copy_docs(ActivityFieldBase.to_data)
    def to_data(self):
        assets_data = {}
        
        image_large = self.image_large
        if (image_large is not None):
            assets_data['large_image'] = image_large
        
        image_small = self.image_small
        if (image_small is not None):
            assets_data['small_image'] = image_small
        
        text_large = self.text_large
        if (text_large is not None):
            assets_data['large_text'] = text_large
        
        text_small = self.text_small
        if (text_small is not None):
            assets_data['small_text'] = text_small
        
        return assets_data
    
    
    @copy_docs(ActivityFieldBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        image_large = self.image_large
        if (image_large is not None):
            repr_parts.append(' image_large=')
            repr_parts.append(repr(image_large))
            field_added = True
        else:
            field_added = False
        
        image_small = self.image_small
        if (image_small is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            repr_parts.append(' image_small=')
            repr_parts.append(repr(image_small))
        
        text_large = self.text_large
        if (text_large is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            repr_parts.append(' text_large=')
            repr_parts.append(repr(text_large))
        
        text_small = self.text_small
        if (text_small is not None):
            if field_added:
                repr_parts.append(',')
            repr_parts.append(' text_small=')
            repr_parts.append(repr(text_small))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    @copy_docs(ActivityFieldBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.image_large != other.image_large:
            return False
        
        if self.image_small != other.image_small:
            return False
        
        if self.text_large != other.text_large:
            return False
        
        if self.text_small != other.text_small:
            return False
        
        return True
    
    
    @copy_docs(ActivityFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        image_large = self.image_large
        if (image_large is not None):
            hash_value ^= hash(image_large)
            hash_value ^= (1 << 0)
        
        image_small = self.image_small
        if (image_small is not None):
            hash_value ^= hash(image_small)
            hash_value ^= (1 << 4)
        
        text_large = self.text_large
        if (text_large is not None):
            hash_value ^= hash(text_large)
            hash_value ^= (1 << 8)
        
        text_small = self.text_small
        if (text_small is not None):
            hash_value ^= hash(text_small)
            hash_value ^= (1 << 12)
        
        return hash_value
    
    
    @copy_docs(ActivityFieldBase.__bool__)
    def __bool__(self):
        image_large = self.image_large
        if (image_large is not None):
            return True
        
        image_small = self.image_small
        if (image_small is not None):
            return True
        
        text_large = self.text_large
        if (text_large is not None):
            return True
        
        text_small = self.text_small
        if (text_small is not None):
            return True
        
        return False



__all__ = ()

from ..color import Color
from ..field_parsers import (
    bool_parser_factory, flag_parser_factory, force_string_parser_factory, int_parser_factory,
    nullable_functional_parser_factor
)
from ..field_putters import (
    bool_optional_putter_factory, flag_optional_putter_factory, force_string_putter_factory, int_putter_factory,
    nullable_functional_optional_putter_factory, string_flag_putter_factory
)
from ..field_validators import (
    bool_validator_factory, entity_id_validator_factory, flag_validator_factory, force_string_validator_factory,
    int_conditional_validator_factory, nullable_entity_conditional_validator_factory, preinstanced_validator_factory
)
from ..permission import Permission
from ..permission.utils import PERMISSION_KEY

from .constants import NAME_LENGTH_MAX, NAME_LENGTH_MIN
from .preinstanced import RoleManagerType

# color

parse_color = flag_parser_factory('color', Color)
put_color_into = flag_optional_putter_factory('color', Color())
validate_color = flag_validator_factory('color', Color)

# manager_id

validate_manager_id = entity_id_validator_factory('manager_id')

# manager_type

validate_manager_type = preinstanced_validator_factory('manager_type', RoleManagerType)

# manager id & type

def parse_manager(data):
    """
    Parses out ``Role.manager_id`` and ``Role.manager_type` fields from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Role data.
    
    Returns
    -------
    manager_type : ``RoleManagerType``
        If the role is manager, defined by what type of entity is the role is managed.
    
    manager_id : `int`
         If the role is managed, then it's manager's id if applicable.
    """    
    while True:
        if not data.get('managed', False):
            manager_type = RoleManagerType.none
            manager_id = 0
            break
        
        try:
            role_tag_data = data['tags']
        except KeyError:
            manager_type = RoleManagerType.unset
            manager_id = 0
            break
        
        try:
            manager_id = role_tag_data['bot_id']
        except KeyError:
            pass
        else:
            manager_type = RoleManagerType.bot
            manager_id = int(manager_id)
            break
        
        if 'premium_subscriber' in role_tag_data:
            manager_type = RoleManagerType.booster
            manager_id = 0
            break
        
        try:
            manager_id = role_tag_data['integration_id']
        except KeyError:
            pass
        else:
            manager_type = RoleManagerType.integration
            manager_id = int(manager_id)
            break
        
        manager_type = RoleManagerType.unknown
        manager_id = 0
        break
    
    return manager_type, manager_id


def put_manager_into(manager, data, defaults):
    """
    Puts the role's manager into the given data.
    
    Parameters
    ----------
    manager : `tuple` of (``RoleManagerType``, `int`)
        The role's manager as a tuple containing it's identifier and type.
    
    data : `dict` of (`str`, `Any`) items
        Role data.
    
    defaults : `bool`
        Whether default field values should be put into `data` as well.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
    """
    manager_type, manager_id = manager
    
    # No goto? LOLK
    while True:
        managed = manager_type is not RoleManagerType.none
        if defaults or managed:
            data['managed'] = managed
        
        if manager_type is RoleManagerType.bot:
            role_manager_data = {'bot_id': str(manager_id)}
        
        elif manager_type is RoleManagerType.booster:
            role_manager_data = {'premium_subscriber': None}
        
        elif manager_type is RoleManagerType.integration:
            role_manager_data = {'integration_id': str(manager_id)}
        
        else:
            break
        
        data['tags'] = role_manager_data
        break
    
    return data


def validate_manager(manager):
    """
    Validates the given manager value.
    
    Parameters
    ----------
    manager : `None`, `tuple` (``RoleManagerType``, `int`)
        Role manager tuple containing it's identifier and type.
    
    Returns
    -------
    manager_type : ``RoleManagerType``
        If the role is manager, defined by what type of entity is the role is managed.
    
    manager_id : `int`
         If the role is managed, then it's manager's id if applicable.
    
    Raises
    ------
    TypeError
        - If `manager`'s type is invalid.
    """
    if manager is None:
        return 0, RoleManagerType.none
    
    if (not isinstance(manager, tuple)) or (len(manager) != 2):
        raise TypeError(
            f'`manager` can be `None` or `tuple` (of length 2 as manager_type & manager_id), got '
            f'{manager.__class__.__name__}; {manager!r}.'
        )
    
    manager_type, manager_id = manager
    manager_type = validate_manager_type(manager_type)
    manager_id = validate_manager_id(manager_id)
    
    return manager_type, manager_id

# mentionable

parse_mentionable = bool_parser_factory('mentionable', False)
put_mentionable_into = bool_optional_putter_factory('mentionable', False)
validate_mentionable = bool_validator_factory('mentionable')

# name

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)

# permissions

parse_permissions = flag_parser_factory(PERMISSION_KEY, Permission)
put_permissions_into = string_flag_putter_factory(PERMISSION_KEY)
validate_permissions = flag_validator_factory(PERMISSION_KEY, Permission)

# position

parse_position = int_parser_factory('position', 0)
put_position_into = int_putter_factory('position')
validate_position = int_conditional_validator_factory(
    'position',
    lambda position : position >= 0,
    '>= 0',
)

# separated

parse_separated = bool_parser_factory('hoist', False)
put_separated_into = bool_optional_putter_factory('hoist', False)
validate_separated = bool_validator_factory('separated')

# unicode_emoji

parse_unicode_emoji = nullable_functional_parser_factor(
    'unicode_emoji', NotImplemented, include = 'create_unicode_emoji'
)
put_unicode_emoji_into = nullable_functional_optional_putter_factory('unicode_emoji', lambda emoji: emoji.unicode)
validate_unicode_emoji = nullable_entity_conditional_validator_factory(
    'unicode_emoji', NotImplemented, lambda emoji: emoji.is_unicode_emoji(), 'unicode emoji', include = 'Emoji'
)
